from __future__ import annotations

import ast
from pathlib import Path

import attrs
import pytest

from rattr.models.symbol import (
    AnyCallInterface,
    CallArguments,
    CallInterface,
    Location,
    Symbol,
)


@attrs.frozen
class _DerivedSymbol(Symbol):
    """Symbol for generic tests."""

    pass


class TestSymbol:
    def test_unable_to_directly_initialise(self):
        with pytest.raises(NotImplementedError):
            Symbol("blah")

    def test_able_to_initialise_sub_classed(self):
        assert _DerivedSymbol(name="blah")


class TestCallInterface:
    def test_all(self):
        interface = CallInterface(
            posonlyargs=["a", "b"],
            args=["arg", "another_arg"],
            kwonlyargs=["kwonly_1", "another_kwonly"],
        )
        assert interface.all == (
            "a",
            "b",
            "arg",
            "another_arg",
            "kwonly_1",
            "another_kwonly",
        )

    def test_all_with_vararg_and_kwarg(self):
        interface = CallInterface(
            posonlyargs=["a", "b"],
            args=["arg", "another_arg"],
            vararg="vararg",
            kwonlyargs=["kwonly_1", "another_kwonly"],
            kwarg="kwarg",
        )
        assert interface.all == (
            "a",
            "b",
            "arg",
            "another_arg",
            "vararg",
            "kwonly_1",
            "another_kwonly",
            "kwarg",
        )

    def test_from_fn_def(self, parse):
        fn = parse(
            """
            def fn(a, b, /, c, d=None, *, e):
                pass
            """
        ).body[0]

        assert CallInterface.from_fn_def(fn) == CallInterface(
            posonlyargs=["a", "b"],
            args=["c", "d"],
            kwonlyargs=["e"],
        )

    def test_from_fn_def_with_vararg_and_kwarg(self, parse):
        fn = parse(
            """
            def fn(a, b, /, c, d=None, *e, f=1, g=2, **h):
                pass
            """
        ).body[0]

        assert CallInterface.from_fn_def(fn) == CallInterface(
            posonlyargs=["a", "b"],
            args=["c", "d"],
            vararg="e",
            kwonlyargs=["f", "g"],
            kwarg="h",
        )

    def test_is_hashable(self):
        assert isinstance(hash(CallInterface()), int)


class TestAnyCallInterface:
    def test_from_fn_def(self):
        with pytest.raises(NotImplementedError):
            AnyCallInterface.from_fn_def(ast.parse("def f(): pass").body[0])

    def test_is_hashable(self):
        assert isinstance(hash(AnyCallInterface()), int)


class TestCallArguments:
    def test_from_call(self):
        _ast = ast.parse("my_func(a, b, c=var)")

        actual = CallArguments.from_call(_ast.body[0].value)
        expected = CallArguments(args=["a", "b"], kwargs={"c": "var"})

        assert actual == expected

    def test_from_call_with_self(self):
        _ast = ast.parse("obj.my_func(a, b, c=var)")

        actual = CallArguments.from_call(_ast.body[0].value, self="self")
        expected = CallArguments(args=["self", "a", "b"], kwargs={"c": "var"})

        assert actual == expected

    def test_is_hashable(self):
        call_arguments = CallArguments(args=["a", "b"], kwargs={"c": "the_real_c"})
        assert isinstance(hash(call_arguments), int)


class TestLocation:
    @pytest.fixture
    def file(self) -> Path:
        return Path("cwd/my_file.py")

    @pytest.fixture
    def required_kwargs(self) -> dict[str, int | None]:
        return {
            "lineno": 0,
            "end_lineno": None,
            "col_offset": 1,
            "end_col_offset": None,
        }

    def test_not_in_file(self, state, required_kwargs):
        with state(current_file=None):
            with pytest.raises(ValueError):
                Location(**required_kwargs)

    def test_with_derived_location(self, state, file, required_kwargs):
        with state(current_file=file):
            assert Location(**required_kwargs).defined_in == file

    def test_with_explicit_location(self, state, file, required_kwargs):
        # This is not the intended interface, but should no break
        with state(current_file=file):
            assert Location(**required_kwargs, file=file).defined_in == file

    def test_is_hashable(self, required_kwargs):
        assert isinstance(hash(Location(**required_kwargs)), int)



from __future__ import annotations

import argparse
from pathlib import Path
from typing import TYPE_CHECKING

from rattr import error
from rattr.cli import _arguments
from rattr.cli._argparse import ArgumentParser
from rattr.cli._types import TomlArgumentType
from rattr.cli._util import get_type_name, multi_paragraph_wrap
from rattr.cli.toml import TOMLDecodeError, parse_project_toml
from rattr.config import Arguments
from rattr.config.util import find_pyproject_toml

if TYPE_CHECKING:
    from typing import Any, NoReturn


TOML_ARGUMENT_NAME_TO_SYS_ARGUMENT_NAME_MAP: dict[str, str] = {
    "exclude-imports": "exclude-import",
}
"""Map from toml argument name to sys argument name.

If absent the name is the same for both the toml arguments and the sys arguments, thus
the expected usage is as such:

>>> # TOML_ARGUMENT_NAME_TO_SYS_ARGUMENT_NAME_MAP = {"saul-of-tarsus": "paul"}
>>> toml_name = "peter"
>>> TOML_ARGUMENT_NAME_TO_SYS_ARGUMENT_NAME_MAP.get(toml_name, toml_name)
"peter"
>>> toml_name = "saul-of-tarsus"
>>> TOML_ARGUMENT_NAME_TO_SYS_ARGUMENT_NAME_MAP.get(toml_name, toml_name)
"paul"
"""

TOML_ARGUMENT_TYPE_MAP: dict[str, TomlArgumentType] = {
    "follow-imports": TomlArgumentType.int,
    "exclude-imports": TomlArgumentType.list_of_strings,
    "exclude": TomlArgumentType.list_of_strings,
    "warning-level": TomlArgumentType.string,
    "collapse-home": TomlArgumentType.flag,
    "truncate-deep-paths": TomlArgumentType.flag,
    "strict": TomlArgumentType.flag,
    "threshold": TomlArgumentType.int,
    "stdout": TomlArgumentType.string,
}
"""The expected type of the arguments in the toml config file.

This is used for:
1. Type checking the provided toml arguments;
2. Correctly converting toml arguments to sys arguments (see TomlArgumentType docs).
"""


def parse_arguments(
    *,
    sys_args: list[str] | None = None,
    project_toml_conf: dict[str, Any] | None = None,
    exit_on_error: bool = True,
) -> Arguments:
    cli_parser = make_cli_parser(exit_on_error=exit_on_error)
    toml_parser = make_toml_parser()

    project_toml_conf = _parse_project_config(
        project_toml_conf,
        _get_toml_override(cli_parser, sys_args=sys_args),
        exit_on_error=exit_on_error,
    )
    toml_arguments = _translate_toml_conf_to_sys_args(project_toml_conf)

    # Parse the arguments from the project toml
    # Then add/override with the arguments from the cli
    arguments = Arguments()
    try:
        toml_parser.parse_args(args=toml_arguments, namespace=arguments)
    except argparse.ArgumentError as argument_error:
        _toml_error(argument_error, exit_on_error=exit_on_error)
    cli_parser.parse_args(args=sys_args, namespace=arguments)

    return arguments


def make_cli_parser(exit_on_error: bool = True) -> ArgumentParser:
    parser = ArgumentParser(
        prog="rattr",
        description=multi_paragraph_wrap(
            """\
            Parse a given Python 3 file to find the attributes used in each function.

            Arguments of type "PATTERN" use Python's regular expression syntax.
            """
        ),
        epilog=multi_paragraph_wrap(
            """\
            Made with ❤️ by Suade Labs.
            """
        ),
        formatter_class=argparse.RawTextHelpFormatter,
        exit_on_error=exit_on_error,
    )

    parser = _arguments.add_version_argument(parser)
    parser = _arguments.add_toml_config_override_argument(parser)
    parser = _arguments.add_common_arguments(parser)
    parser = _arguments.add_target_file_argument(parser)

    return parser


def make_toml_parser() -> ArgumentParser:
    parser = ArgumentParser(exit_on_error=False)
    parser = _arguments.add_common_arguments(parser)

    return parser


def _get_toml_override(
    cli_parser: ArgumentParser,
    *,
    sys_args: list[str] | None = None,
) -> Path | None:
    """Return the toml config override from the CLI if given, otherwise `None`."""
    cli_arguments = cli_parser.parse_args(args=sys_args, namespace=Arguments())
    pyproject_toml_override = cli_arguments.pyproject_toml_override

    if pyproject_toml_override and pyproject_toml_override.is_file():
        return pyproject_toml_override

    return None


def _toml_error(exc: Exception, *, exit_on_error: bool) -> NoReturn:
    if exit_on_error:
        error.fatal(f"error parsing project toml: {error}")
    raise exc


def _parse_project_config(
    input_config: dict[str, Any] | None,
    project_toml_override: Path | None,
    *,
    exit_on_error: bool = True,
) -> dict[str, Any]:
    pyproject_toml = find_pyproject_toml()

    # Allow the use of an explicit config
    # This is helpful for testing and for making rattr derived projects
    if not input_config:
        try:
            conf = parse_project_toml(pyproject_toml, project_toml_override)
        except TOMLDecodeError as toml_decode_error:
            _toml_error(toml_decode_error, exit_on_error=exit_on_error)
    else:
        conf = input_config

    try:
        conf = _validate_toml_config(conf)
    except argparse.ArgumentError as argument_error:
        _toml_error(argument_error, exit_on_error=exit_on_error)

    return conf


def _validate_toml_config(conf: dict[str, Any]) -> dict[str, Any]:
    _type_error: str = (
        "{arg!r} expects type {expected_type}, got {value!r} of type {actual_type}"
    )

    # Prune unrecognised arguments
    toml_supported_arguments = TOML_ARGUMENT_TYPE_MAP.keys()
    conf = {k: v for k, v in conf.items() if k in toml_supported_arguments}

    # Validate types
    for arg, value in conf.items():
        expected_type = TOML_ARGUMENT_TYPE_MAP[arg]

        if not expected_type.is_valid(value):
            _error = _type_error.format(
                arg=arg,
                value=value,
                expected_type=str(expected_type),
                actual_type=get_type_name(value),
            )
            raise argparse.ArgumentError(None, _error)

    return conf


def _translate_toml_conf_to_sys_args(toml_conf: dict[str, Any]) -> list[str]:
    """Return the toml conf translated to `sys.argv` style list.

    >>> _translate_toml_conf_to_sys_args(
    ...     {"threshold": 10, "warning-level": "all", "exclude": ["a", "b"]}
    ... )
    ["--threshold", "0", "--warning-level", "all", "--exclude", "a", "--exclude", "b"]
    """
    toml_sys_args = []

    for k, v in toml_conf.items():
        if len(k) > 1:
            arg_name = f"--{TOML_ARGUMENT_NAME_TO_SYS_ARGUMENT_NAME_MAP.get(k, k)}"
        else:
            arg_name = f"-{k}"

        if isinstance(v, bool):
            toml_sys_args += [arg_name] if v else []
        elif isinstance(v, str) or isinstance(v, int) or isinstance(v, float):
            toml_sys_args += [arg_name, f"{v}"]
        elif isinstance(v, list):
            for arg_value in v:
                toml_sys_args += [arg_name, f"{arg_value}"]

    return toml_sys_args



"""End-to-end regression tests."""

from __future__ import annotations

import json
from pathlib import Path
from typing import TYPE_CHECKING

import pytest

from rattr.analyser.file import parse_and_analyse_file
from rattr.cli import parse_arguments
from rattr.config import Config
from rattr.models.util import OutputIrs, deserialise, serialise, serialise_irs
from rattr.results import generate_results_from_ir
from tests.regression.shared import (
    assert_actual_and_expected_have_the_same_functions_irs,
    assert_actual_and_expected_have_the_same_functions_results,
    code_dir,
    code_files,
    ir_files,
    results_files,
    results_function_diffs,
)

if TYPE_CHECKING:
    from typing import Any

    from tests.shared import OsDependentPathFn


@pytest.mark.parametrize(
    ("code_file,results_file"),
    zip(code_files, results_files),
    ids=[str(f.relative_to(code_dir)) for f in code_files],
)
def test_run_e2e_regression_tests_for_results(
    code_file: Path,
    results_file: Path,
):
    # TODO
    #   Make this more end-to-end-y (that is, use `main(...))
    #   This should suffice for now as `main(...)` is not particularly practical for
    #   this test at present, but will be shortly, at which point we can make a
    #   small refactor here.

    # Parse expected results
    expected_results: dict[str, Any] = json.loads(results_file.read_text())

    # Setup simulated cli arguments and state
    config = Config()
    config.arguments = parse_arguments(
        sys_args=[
            "--collapse-home",
            *("--warning", "none"),
            *("--threshold", "0"),
            *("--stdout", "results"),
            str(code_file),
        ],
    )

    # Equivalent to main function
    file_ir, import_irs, _ = parse_and_analyse_file()
    actual_results = generate_results_from_ir(target_ir=file_ir, import_irs=import_irs)

    if not actual_results:
        pytest.skip(f"no results for {code_file}")

    # Assert unordered (semantic) equality
    assert_actual_and_expected_have_the_same_functions_results(
        actual_results,
        expected_results,
    )

    full_file_diff = {
        fn: results_function_diffs(fn_results, expected_results[fn])
        for fn, fn_results in actual_results.items()
    }

    diff_is_empty = all(
        diff == []
        for _, fn_diffs in full_file_diff.items()
        for _, diff in fn_diffs.items()
    )
    diff_error = (
        f"actual and expected results have the following diff:\n"
        f"{json.dumps(full_file_diff, indent=4)}"
    )

    assert diff_is_empty, diff_error


@pytest.mark.update_expected_results
@pytest.mark.parametrize(
    ("code_file,results_file"),
    zip(code_files, results_files),
    ids=[str(f.relative_to(code_dir)) for f in code_files],
)
def test_update_expected_results(code_file: Path, results_file: Path):
    # Setup simulated cli arguments and state
    config = Config()
    config.arguments = parse_arguments(
        sys_args=[
            *("--warning", "all"),
            *("--threshold", "0"),
            *("--stdout", "results"),
            str(code_file),
        ]
    )

    file_ir, import_irs, _ = parse_and_analyse_file()
    results = generate_results_from_ir(target_ir=file_ir, import_irs=import_irs)

    results_file.write_text(serialise(results, indent=4))


@pytest.mark.posix
@pytest.mark.cpython
@pytest.mark.python_3_9
@pytest.mark.parametrize(
    ("code_file,ir_file"),
    zip(code_files, ir_files),
    ids=[str(f.relative_to(code_dir)) for f in code_files],
)
def test_run_e2e_regression_tests_for_ir(
    code_file: Path,
    ir_file: Path,
    os_dependent_path: OsDependentPathFn,
):
    # TODO See todo in test_run_e2e_regression_tests_for_results

    # Parse expected results
    expected_irs = deserialise(ir_file.read_text(), type=OutputIrs)

    # Setup simulated cli arguments and state
    config = Config()
    config.arguments = parse_arguments(
        sys_args=[
            "--collapse-home",
            *("--warning", "none"),
            *("--threshold", "0"),
            *("--stdout", "ir"),
            str(code_file),
        ],
    )

    # Equivalent to main function
    file_ir, import_irs, _ = parse_and_analyse_file()
    _ = generate_results_from_ir(target_ir=file_ir, import_irs=import_irs)

    # Ensure filename is the same in CI
    actual_tests_dir = Path(__file__).parents[1]
    expected_tests_dir = Path("/home/suade/rattr/feature/tests")
    filename = expected_tests_dir / code_file.relative_to(actual_tests_dir)

    irs = OutputIrs(
        import_irs=import_irs,
        target_ir={
            "filename": os_dependent_path(str(filename)),
            "ir": file_ir,
        },
    )

    if not file_ir and not import_irs:
        pytest.skip(f"no irs for {code_file}")

    # Assert unordered (semantic) equality
    assert irs.import_irs.keys() == expected_irs.import_irs.keys()
    for mod in irs.import_irs.keys():
        assert_actual_and_expected_have_the_same_functions_irs(
            irs.import_irs[mod],
            expected_irs.import_irs[mod],
        )
    assert_actual_and_expected_have_the_same_functions_irs(
        irs.target_ir["ir"],
        expected_irs.target_ir["ir"],
    )

    assert irs.import_irs == expected_irs.import_irs
    assert irs.target_ir["filename"] == expected_irs.target_ir["filename"]
    assert irs.target_ir["ir"] == expected_irs.target_ir["ir"]


@pytest.mark.update_expected_irs
@pytest.mark.parametrize(
    ("code_file,ir_files"),
    zip(code_files, ir_files),
    ids=[str(f.relative_to(code_dir)) for f in code_files],
)
def test_update_expected_ir(code_file: Path, ir_files: Path):
    # Setup simulated cli arguments and state
    config = Config()
    config.arguments = parse_arguments(
        sys_args=[
            *("--warning", "all"),
            *("--threshold", "0"),
            *("--stdout", "ir"),
            str(code_file),
        ]
    )

    file_ir, import_irs, _ = parse_and_analyse_file()
    _ = generate_results_from_ir(target_ir=file_ir, import_irs=import_irs)

    serialised = serialise_irs(
        target_name=str(code_file),
        target_ir=file_ir,
        import_irs=import_irs,
    )
    ir_files.write_text(serialised)



from __future__ import annotations

import ast
import re
import sys
from pathlib import Path
from typing import TYPE_CHECKING
from unittest import mock

import pytest

from rattr import error
from rattr.analyser.exc import RattrResultsError
from rattr.analyser.util import (
    DictChanges,
    assignment_is_one_to_one,
    class_in_rhs,
    get_annotation,
    get_assignment_targets,
    get_attrname,
    get_basename,
    get_basename_fullname_pair,
    get_contained_walruses,
    get_decorator_name,
    get_dynamic_name,
    get_first_argument_name,
    get_fullname,
    get_function_body,
    get_function_form,
    get_namedtuple_attrs_from_call,
    get_xattr_obj_name_pair,
    has_affect,
    has_annotation,
    is_blacklisted_module,
    is_call_to,
    is_excluded_name,
    is_in_builtins,
    is_list_of_call_specs,
    is_list_of_names,
    is_method_on_cast,
    is_method_on_constant,
    is_method_on_primitive,
    is_name,
    is_pip_module,
    is_relative_import,
    is_set_of_names,
    is_starred_import,
    is_stdlib_module,
    lambda_in_rhs,
    namedtuple_in_rhs,
    parse_annotation,
    parse_rattr_results_from_annotation,
    unravel_names,
    validate_rattr_results,
    walrus_in_rhs,
)
from rattr.models.context import Context, compile_root_context
from rattr.models.symbol import Call, CallArguments, CallInterface, Class, Name

if TYPE_CHECKING:
    from collections.abc import Iterator

    from rattr.analyser.types import RattrResults
    from tests.shared import ArgumentsFn, StateFn


@pytest.fixture(autouse=True)
def __set_current_file(state: StateFn) -> Iterator[None]:
    with state(current_file=Path(__file__)):
        yield


class TestUtil:
    def test_get_basename_full_name_pair(self):
        # Simple nameable
        nameable = ast.parse("name").body[0].value
        expected = ("name", "name")
        assert get_basename_fullname_pair(nameable) == expected

        # Complex/nested strictly nameable
        nameable = ast.parse("a.b[0].attr").body[0].value
        expected = ("a", "a.b[].attr")

        assert get_basename_fullname_pair(nameable) == expected

    def test_get_basename_full_name_pair_safety(self):
        # Safety off
        nameable = ast.parse("(a, b)[0].attr").body[0].value

        with pytest.raises(error.RattrLiteralInNameable):
            get_basename_fullname_pair(nameable)

        # Safety on
        nameable = ast.parse("(a, b)[0].attr").body[0].value
        expected = ("@Tuple", "@Tuple[].attr")

        assert get_basename_fullname_pair(nameable, safe=True) == expected

    def test_get_basename_full_name_pair_coverage(self, constant: str):
        # Name
        nameable = ast.parse("ast_name").body[0].value
        expected = ("ast_name", "ast_name")
        assert get_basename_fullname_pair(nameable) == expected

        # Attribute
        nameable = ast.parse("ast_name.attr").body[0].value
        expected = ("ast_name", "ast_name.attr")
        assert get_basename_fullname_pair(nameable) == expected

        # Subscript
        nameable = ast.parse("ast_name[0]").body[0].value
        expected = ("ast_name", "ast_name[]")
        assert get_basename_fullname_pair(nameable) == expected

        # Call
        nameable = ast.parse("ast_name(arg_one, arg_two)").body[0].value
        expected = ("ast_name", "ast_name()")
        assert get_basename_fullname_pair(nameable) == expected

        # Starred
        nameable = ast.parse("*ast_name").body[0].value
        expected = ("ast_name", "*ast_name")
        assert get_basename_fullname_pair(nameable) == expected

        # UnaryOp
        nameable = ast.parse("(-a)").body[0].value
        expected = ("@UnaryOp", "@UnaryOp")
        with pytest.raises(error.RattrUnaryOpInNameable):
            get_basename_fullname_pair(nameable)
        assert get_basename_fullname_pair(nameable, safe=True) == expected

        # BinOp
        nameable = ast.parse("(a + b)").body[0].value
        expected = ("@BinOp", "@BinOp")
        with pytest.raises(error.RattrBinOpInNameable):
            get_basename_fullname_pair(nameable)
        assert get_basename_fullname_pair(nameable, safe=True) == expected

        # NameConstant
        nameable = ast.parse("True").body[0].value
        expected = (constant, constant)
        with pytest.raises(error.RattrConstantInNameable):
            get_basename_fullname_pair(nameable)
        assert get_basename_fullname_pair(nameable, safe=True) == expected

        # Literal
        nameable = ast.parse("(1).to_bytes()").body[0].value
        expected = (constant, f"{constant}.to_bytes()")
        with pytest.raises(error.RattrConstantInNameable):
            get_basename_fullname_pair(nameable)
        assert get_basename_fullname_pair(nameable, safe=True) == expected

        # Comprehension
        nameable = ast.parse("[a for a in list_of_as][0]").body[0].value
        expected = ("@ListComp", "@ListComp[]")
        with pytest.raises(error.RattrComprehensionInNameable):
            get_basename_fullname_pair(nameable)
        assert get_basename_fullname_pair(nameable, safe=True) == expected

    def test_get_basename_full_name_pair_regression(self, parse):
        # Nameable is a getattr, etc call #1
        nameable = ast.parse("getattr(getattr(a, 'inner'), 'outer')").body[0].value
        expected = ("getattr", "a.inner.outer")
        assert get_basename_fullname_pair(nameable) == expected

        # Nameable is a getattr, etc call #2
        nameable = ast.parse("getattr(a, 'inner').outer").body[0].value
        expected = ("getattr", "a.inner.outer")
        assert get_basename_fullname_pair(nameable) == expected

    def test_get_basename(self):
        # Simple strictly nameable
        nameable = ast.parse("name").body[0].value
        expected = "name"

        assert get_basename(nameable) == expected

        # Complex strictly nameable
        nameable = ast.parse("a.b[0].attr").body[0].value
        expected = "a"

        assert get_basename(nameable) == expected

        # Complex strictly nameable
        nameable = ast.parse("(a, b)[0].attr").body[0].value
        expected = "@Tuple"

        with pytest.raises(error.RattrLiteralInNameable):
            assert get_basename(nameable) == expected

    def test_unravel_names(self):
        # Simple
        ravelled_names = ast.parse("a, b = 1, 2").body[0].targets[0]
        expected = ["a", "b"]

        assert list(unravel_names(ravelled_names)) == expected

        # Simple
        ravelled_names = ast.parse("(a, b), c = [1, 2], 3").body[0].targets[0]
        expected = ["a", "b", "c"]

        assert list(unravel_names(ravelled_names)) == expected

        # Complex
        ravelled_names = ast.parse("(a, b), c, d.e = 1, 2, 3, 4").body[0].targets[0]
        expected = ["a", "b", "c", "d"]

        assert list(unravel_names(ravelled_names)) == expected

    def test_unravel_names_by_fullname(self):
        # Simple
        ravelled_names = ast.parse("a, b = 1, 2").body[0].targets[0]
        expected = ["a", "b"]

        assert list(unravel_names(ravelled_names, get_name=get_fullname)) == expected

        # Simple
        ravelled_names = ast.parse("(a, b), c = [1, 2], 3").body[0].targets[0]
        expected = ["a", "b", "c"]

        assert list(unravel_names(ravelled_names, get_name=get_fullname)) == expected

        # Complex
        ravelled_names = ast.parse("(a, b), c, d.e = 1, 2, 3, 4").body[0].targets[0]
        expected = ["a", "b", "c", "d.e"]

        assert list(unravel_names(ravelled_names, get_name=get_fullname)) == expected

    def test_is_call_to(self):
        not_a_call = ast.parse("def f(): pass").body[0]
        with pytest.raises(TypeError):
            is_call_to("blah", not_a_call)

        call = ast.parse("fn(a, b, c)").body[0].value
        assert is_call_to("different_func", call) is False
        assert is_call_to("fn", call) is True

        call = ast.parse("outer(inner(a, b, c))").body[0].value
        assert is_call_to("different_func", call) is False
        assert is_call_to("inner", call) is False
        assert is_call_to("outer", call) is True

        call = ast.parse("obj.method(a, b, c)").body[0].value
        assert is_call_to("different_func", call) is False
        assert is_call_to("object.method", call) is False

        call = ast.parse("fn(a, b, c).on_result()").body[0].value
        assert is_call_to("different_func", call) is False
        assert is_call_to("on_result", call) is False
        assert is_call_to("fn", call) is False
        assert is_call_to("fn", call.func.value) is True

    def test_has_affect(self, builtins):
        _has_affect = {
            "hasattr",
            "getattr",
            "setattr",
            "delattr",
        }

        for b in builtins.difference(_has_affect):
            assert not has_affect(b)

        for b in _has_affect:
            assert has_affect(b)

        with pytest.raises(ValueError):
            has_affect("not_a_builtin")

    def test_get_xattr_obj_name_pair(self, capfd):
        # Simple
        xattr = ast.parse("getattr(a, 'b')").body[0].value
        expected = ("a", "b")

        assert get_xattr_obj_name_pair("getattr", xattr) == expected

        # Nested
        xattr = ast.parse("getattr(getattr(a, 'b'), 'c')").body[0].value
        expected = ("a.b", "c")

        assert get_xattr_obj_name_pair("getattr", xattr) == expected

        # Complex
        xattr = ast.parse("getattr(getattr(a.b[0], 'c'), 'd')").body[0].value
        expected = ("a.b[].c", "d")

        assert get_xattr_obj_name_pair("getattr", xattr) == expected

        # Warn: non string-literal name
        xattr = ast.parse("getattr(a, some_string_variable)").body[0].value
        expected = ("a", "<some_string_variable>")

        assert get_xattr_obj_name_pair("getattr", xattr, True) == expected

        _, stderr = capfd.readouterr()
        assert "error" in stderr
        assert "expects name to be a string literal" in stderr

        # Illegal: mixed nested
        xattr = ast.parse("getattr(hasattr(a, 'b'), 'c')").body[0].value

        with mock.patch("sys.exit") as _exit:
            get_xattr_obj_name_pair("getattr", xattr, True)

        _, stderr = capfd.readouterr()

        assert "object must be a name or a call to 'getattr'" in stderr
        assert _exit.call_count == 1

        # Illegal: expression
        xattr = ast.parse("getattr(a or b, 'c')").body[0].value

        with pytest.raises(TypeError):
            get_xattr_obj_name_pair("getattr", xattr)

        # Illegal: exprssion nested
        xattr = ast.parse("getattr(getattr(a1 or a2, 'b'), 'c')").body[0].value

        with pytest.raises(TypeError):
            get_xattr_obj_name_pair("getattr", xattr)

    def test_get_decorator_name(self, parse):
        # Works with ast.Name decorator
        _ast = parse(
            """
            @name_decorator
            def fn(bob: YourUncle):
                return isinstance(bob, YourUncle)
            """
        )
        decorator = _ast.body[0].decorator_list[0]
        assert get_decorator_name(decorator) == "name_decorator"

        # Works with ast.Call decorator
        _ast = parse(
            """
            @called_decorator("some decorator arg")
            @called_decorator_no_args()
            def fn(bob: YourUncle):
                return isinstance(bob, YourUncle)
            """
        )
        decorator_one = _ast.body[0].decorator_list[0]
        decorator_two = _ast.body[0].decorator_list[1]
        assert get_decorator_name(decorator_one) == "called_decorator"
        assert get_decorator_name(decorator_two) == "called_decorator_no_args"

        # Error otherwise
        decorator = parse(
            """
            x = 4       # <result>.body[0] will be ast.Assign
            """
        ).body[0]
        with pytest.raises(TypeError):
            get_decorator_name(decorator)

    def test_get_first_argument_name(self, parse):
        _ast = parse(
            """
            def func_call_no_args():
                pass
            def func_call_one_arg(arg):
                pass
            def func_call_many_args(arg_one, arg_two, arg_three):
                pass
            """
        )

        # Return empty string when there are no args
        args = _ast.body[0].args
        assert get_first_argument_name(args) == ""

        # Return first arg of one args
        args = _ast.body[1].args
        assert get_first_argument_name(args) == "arg"

        # Return first arg of many args
        args = _ast.body[2].args
        assert get_first_argument_name(args) == "arg_one"

    def test_has_annotation(self, parse):
        _ast = parse(
            """
            def no_decorators(a, b, c):
                pass

            @decorator
            def single_decorated(a, b, c):
                pass

            @decorator_one
            @decorator_two
            def multi_decorated(a, b, c):
                pass

            @called_decorator()
            def call_decorated(a, b, c):
                pass

            @nested.attribute.decorator
            def attribute_decorated(a, b, c):
                pass

            @nested.attribute.call.decorator()
            def call_attribute_decorated(a, b, c):
                pass
            """
        )

        # No decorators
        fn_def = _ast.body[0]
        assert not has_annotation("absent", fn_def)

        # Decorated
        fn_def = _ast.body[1]
        assert not has_annotation("absent", fn_def)
        assert has_annotation("decorator", fn_def)

        # Multi
        fn_def = _ast.body[2]
        assert not has_annotation("absent", fn_def)
        assert has_annotation("decorator_one", fn_def)
        assert has_annotation("decorator_two", fn_def)

        # Called
        fn_def = _ast.body[3]
        assert not has_annotation("absent", fn_def)
        assert has_annotation("called_decorator", fn_def)

        # Attribute
        fn_def = _ast.body[4]
        assert not has_annotation("absent", fn_def)
        assert has_annotation("decorator", fn_def)

        # Called attribute
        fn_def = _ast.body[5]
        assert not has_annotation("absent", fn_def)
        assert has_annotation("decorator", fn_def)

    def test_get_annotation(self, parse, capfd):
        _ast = parse(
            """
            def no_decorators(a, b, c):
                pass

            @present
            @present
            def duplicated(a, b, c):
                pass

            @present
            def normal(a, b, c):
                pass

            @present()
            def called(a, b, c):
                pass

            @present("value", "another value")
            def called_with_args(a, b, c):
                pass

            @a.b.c.present("value", "another value")
            def called_attr_with_args(a, b, c):
                pass
            """
        )

        # No decorators
        fn_def = _ast.body[0]
        assert get_annotation("absent", fn_def) is None

        # Duplicated
        fn_def = _ast.body[1]
        assert get_annotation("absent", fn_def) is None

        _, stderr = capfd.readouterr()  # clear stderr buffer
        with mock.patch("sys.exit") as _exit:
            get_annotation("present", fn_def)
        _, stderr = capfd.readouterr()

        assert "duplicated annotation" in stderr
        assert _exit.call_count == 1

        # Normal
        fn_def = _ast.body[2]
        assert get_annotation("absent", fn_def) is None

        expected = "Name(id='present', ctx=Load())"
        assert ast.dump(get_annotation("present", fn_def)) == expected

        # Called
        fn_def = _ast.body[3]
        assert get_annotation("absent", fn_def) is None

        expected = "Call(func=Name(id='present', ctx=Load()), " "args=[], keywords=[])"
        assert ast.dump(get_annotation("present", fn_def)) == expected

        # Called w/ args
        fn_def = _ast.body[4]
        assert get_annotation("absent", fn_def) is None

        if sys.version_info.major == 3 and sys.version_info.minor == 7:
            expected = (
                "Call(func=Name(id='present', ctx=Load()), "
                "args=[Str(s='value'), Str(s='another value')], keywords=[])"
            )
        elif sys.version_info.major == 3 and sys.version_info.minor == 8:
            expected = (
                "Call(func=Name(id='present', ctx=Load()), "
                "args=[Constant(value='value', kind=None), "
                "Constant(value='another value', kind=None)], keywords=[])"
            )
        else:
            expected = (
                "Call(func=Name(id='present', ctx=Load()), "
                "args=[Constant(value='value'), "
                "Constant(value='another value')], keywords=[])"
            )
        assert ast.dump(get_annotation("present", fn_def)) == expected

        # Called Attr w/ args
        fn_def = _ast.body[5]
        assert get_annotation("absent", fn_def) is None

        if sys.version_info.major == 3 and sys.version_info.minor == 7:
            expected = (
                "Call(func=Attribute(value=Attribute(value=Attribute(value="
                "Name(id='a', ctx=Load()), attr='b', ctx=Load()), attr='c', "
                "ctx=Load()), attr='present', ctx=Load()), "
                "args=[Str(s='value'), Str(s='another value')], keywords=[])"
            )
        elif sys.version_info.major == 3 and sys.version_info.minor == 8:
            expected = (
                "Call(func=Attribute(value=Attribute(value=Attribute(value="
                "Name(id='a', ctx=Load()), attr='b', ctx=Load()), attr='c', "
                "ctx=Load()), attr='present', ctx=Load()), "
                "args=[Constant(value='value', kind=None), "
                "Constant(value='another value', kind=None)], keywords=[])"
            )
        else:
            expected = (
                "Call(func=Attribute(value=Attribute(value=Attribute(value="
                "Name(id='a', ctx=Load()), attr='b', ctx=Load()), attr='c', "
                "ctx=Load()), attr='present', ctx=Load()), "
                "args=[Constant(value='value'), "
                "Constant(value='another value')], keywords=[])"
            )
        assert ast.dump(get_annotation("present", fn_def)) == expected

    def test_parse_annotation(self, parse):
        _ast = parse(
            """
            @name
            def fn(a, b, c):
                pass

            @call()
            def fn(a, b, c):
                pass

            @call("simple and evaluateable")
            def fn(a, b, c):
                pass

            @call("arg", kwarg="value")
            def fn(a, b, c):
                pass

            @call([1, 2, 3], gets={"more", "complex", "test"})
            def fn(a, b, c):
                pass

            @call(d={"a": 1, "b": 2})
            def fn(a, b, c):
                pass

            # Illegals

            @call(jeremy, james, hammond, show=TopGear)
            def fn(a, b, c):
                pass

            @call([1, 2, illegal])
            def fn(a, b, c):
                pass
            """
        )

        the_empty_result = ([], {})

        # No value, name
        fn_def = _ast.body[0]
        assert parse_annotation("absent", fn_def) == the_empty_result

        # No value, call
        fn_def = _ast.body[1]
        assert parse_annotation("absent", fn_def) == the_empty_result
        assert parse_annotation("call", fn_def) == the_empty_result

        # Simple
        fn_def = _ast.body[2]
        assert parse_annotation("absent", fn_def) == the_empty_result
        assert parse_annotation("call", fn_def) == (
            ["simple and evaluateable"],
            {},
        )

        # Keyword
        fn_def = _ast.body[3]
        assert parse_annotation("absent", fn_def) == the_empty_result
        assert parse_annotation("call", fn_def) == (
            ["arg"],
            {"kwarg": "value"},
        )

        # Complex
        fn_def = _ast.body[4]
        assert parse_annotation("absent", fn_def) == the_empty_result
        assert parse_annotation("call", fn_def) == (
            [
                [1, 2, 3],
            ],
            {"gets": set(["more", "complex", "test"])},
        )

        # Dict
        fn_def = _ast.body[5]
        assert parse_annotation("absent", fn_def) == the_empty_result
        assert parse_annotation("call", fn_def) == (
            [],
            {
                "d": {"a": 1, "b": 2},
            },
        )

        # Illegal: non-compile time evaluated
        fn_def = _ast.body[6]
        assert parse_annotation("absent", fn_def) == the_empty_result

        with mock.patch("sys.exit") as _exit:
            parse_annotation("call", fn_def)
        assert _exit.call_count == 4  # exit called once per argument

        # Illegal: nested non-compile time evaluated
        fn_def = _ast.body[7]
        assert parse_annotation("absent", fn_def) == the_empty_result

        with mock.patch("sys.exit") as _exit:
            parse_annotation("call", fn_def)
        assert _exit.call_count == 1

    def test_is_name(self):
        assert is_name("var")
        assert is_name("var.attr")

        assert is_name("*var.attr")
        assert is_name("*var.attr")

        assert is_name("*var.mth().res[].attr")
        assert is_name("a_long_var.with_underscores.in_it")
        assert is_name("var_with_2_numers_in_it_1")
        assert is_name("_protected")
        assert is_name("__private")
        assert is_name("ClassName")

        assert is_name("@BinOp")
        assert is_name("@BinOp.attr")

        assert not is_name(1234)
        assert not is_name("1234")
        assert not is_name(".var")
        assert not is_name("1_var")

        assert not is_name("")

    @pytest.mark.parametrize(
        "input, expected",
        [
            # the empty set
            (set(), True),
            # expected usage
            ({"a", "b", "c"}, True),
            ({"var.a", "var_b", "*c"}, True),
            # other empty iterable
            ([], False),
            ({}, False),
            # iterable, but wrongly typed
            ([1, 2, 3, 4], False),
            (["a", "b", "c"], False),
            # set with non-names
            ({"a", "b", "1111"}, False),
            # not an iterable
            (1234, False),
        ],
    )
    def test_is_set_of_names(self, input: object, expected: bool):
        assert is_set_of_names(input) == expected

    @pytest.mark.parametrize(
        "input, expected",
        [
            # the empty list
            ([], True),
            # expected usage
            (["a", "b", "c"], True),
            (["var.a", "var_b", "*c"], True),
            # other empty iterable
            (set(), False),
            ({}, False),
            # iterable, but wrongly typed
            ([1, 2, 3, 4], False),
            ({"a", "b", "c"}, False),
            # set with non-names
            ({"a", "b", "1111"}, False),
            # not an iterable
            (1234, False),
        ],
    )
    def test_is_list_of_names(self, input: object, expected: bool):
        assert is_list_of_names(input) == expected

    @pytest.mark.parametrize(
        "input, expected",
        testcases := [
            # typical call specs
            (([], {}), True),
            ((["a"], {}), True),
            ((["a", "b", "c"], {}), True),
            (([], {"a": "A"}), True),
            (([], {"a": "A", "b": "B"}), True),
            ((["a"], {"b": "B"}), True),
            ((["a", "b", "c"], {"d": "D"}), True),
            ((["a", "b", "c"], {"d": "D", "e": "E"}), True),
            # look like call specs, but wrongly typed
            ((["1", "b"], {"c": "d"}), False),
            ((["a", "b"], {"1": "d"}), False),
            ((["a", "b"], {"c": "1"}), False),
            (([111, "b"], {"c": "d"}), False),
            ((["a", "b"], {111: "d"}), False),
            ((["a", "b"], {"c": 111}), False),
            # entirely incorrect
            (1234, False),
            ("abcd", False),
            ((["a", "b"],), False),
        ],
        ids=[i for i, _ in enumerate(testcases)],
    )
    def test_is_call_spec(self, input: object, expected: bool):
        assert is_list_of_call_specs([("my_function", input)]) == expected

    @pytest.mark.parametrize(
        "inputs, expected",
        testcases := [
            # correct
            (
                [
                    ([], {}),
                    (["a"], {}),
                    ([], {"a": "A", "b": "B"}),
                    (["a", "b", "c"], {"d": "D", "e": "E"}),
                ],
                True,
            ),
            # mostly correct but one is wrong
            (
                [
                    ([], {}),
                    (["a"], {}),
                    ("NAUGHTY"),
                    (["a", "b", "c"], {"d": "D", "e": "E"}),
                ],
                False,
            ),
        ],
        ids=[i for i, _ in enumerate(testcases)],
    )
    def test_is_list_of_call_specs(self, inputs: object, expected: bool):
        assert is_list_of_call_specs([("func", input) for input in inputs]) == expected

    @pytest.mark.parametrize(
        "input, error",
        testcases := [
            (
                {
                    "gets": ["fail here"],
                    "sets": set(),
                    "dels": set(),
                    "calls": [],
                },
                "'rattr_results' expects a set[Identifier] for 'gets', where "
                "Identifier is a type alias for str",
            ),
            (
                {
                    "gets": set(),
                    "sets": set(),
                    "dels": set(),
                    "calls": [("fail here", ...)],
                },
                "'rattr_results' expects 'calls' to be a "
                "list[tuple[TargetName, tuple[list[PositionalArgumentName], "
                "dict[KeywordArgumentName, LocalIdentifier]]]]; "
                "where TargetName, PositionalArgumentName, KeywordArgumentName, and "
                "LocalIdentifier are type aliases for str",
            ),
        ],
        ids=[i for i, _ in enumerate(testcases)],
    )
    def test_validate_rattr_results(self, input: RattrResults, error: str):
        with pytest.raises(RattrResultsError, match=re.escape(error)):
            validate_rattr_results(input)

    def test_parse_rattr_results_from_annotation(self, parse):
        ast_module = parse(
            """
            @rattr_results(
                sets={"a"},
                gets={"b"},
                calls=[("c()", ([], {}))],
                dels={"d"}
            )
            def fn():
                pass

            @rattr_results(sets={"a", "b"})
            def fn():
                pass

            # Illegal

            @rattr_results("posarg")
            def fn():
                pass

            @rattr_results(invalid_key=1)
            def fn():
                pass

            @rattr_results(sets="invalid type")
            def fn():
                pass
            """
        )
        context = compile_root_context(ast_module)

        # Can set values
        fn_def = ast_module.body[0]
        expected = {
            "sets": {Name("a")},
            "gets": {Name("b")},
            "calls": {Call("c", args=CallArguments())},
            "dels": {Name("d")},
        }
        assert parse_rattr_results_from_annotation(fn_def, context=context) == expected

        # Defaults
        fn_def = ast_module.body[1]
        expected = {
            "sets": {Name("a"), Name("b")},
            "gets": set(),
            "calls": set(),
            "dels": set(),
        }
        assert parse_rattr_results_from_annotation(fn_def, context=context) == expected

        # Illegal: has pos arg
        fn_def = ast_module.body[2]
        with mock.patch("sys.exit") as _exit:
            parse_rattr_results_from_annotation(fn_def, context=context)
        assert _exit.call_count == 1

        # Illegal: has invalid key
        fn_def = ast_module.body[3]
        with mock.patch("sys.exit") as _exit:
            parse_rattr_results_from_annotation(fn_def, context=context)
        assert _exit.call_count == 1

        # Illegal: has invalid type
        fn_def = ast_module.body[4]
        with mock.patch("sys.exit") as _exit:
            parse_rattr_results_from_annotation(fn_def, context=context)
        assert _exit.call_count == 1

    def test_parse_rattr_results_from_annotation_complex(self, parse):
        _ast = parse(
            """
            @rattr_results(
                sets={"s.attr"},
                gets={"g.mth().res[]", "*g.attr"},
                dels={"del_me"}
            )
            def fn():
                pass

            @rattr_results(calls=[
                ("fn_a()", ([], {})),
                ("fn_b()", (["a", "b"], {})),
                ("fn_c()", (["a", ], {"c": "@Str"})),
            ])
            def fn():
                pass
            """
        )
        _ctx = compile_root_context(_ast)

        # Complex names
        fn_def = _ast.body[0]
        expected = {
            "sets": {Name("s.attr", "s")},
            "gets": {Name("g.mth().res[]", "g"), Name("*g.attr", "g")},
            "dels": {Name("del_me")},
            "calls": set(),
        }
        assert parse_rattr_results_from_annotation(fn_def, context=_ctx) == expected

        # Complex calls
        fn_def = _ast.body[1]
        expected = {
            "sets": set(),
            "gets": set(),
            "dels": set(),
            "calls": {
                Call(name="fn_a", args=CallArguments()),
                Call(name="fn_b", args=CallArguments(args=("a", "b"))),
                Call(
                    name="fn_c",
                    args=CallArguments(args=("a",), kwargs={"c": "@Str"}),
                ),
            },
        }
        assert parse_rattr_results_from_annotation(fn_def, context=_ctx) == expected

    def test_parse_rattr_results_from_annotation_follow_call(self, parse):
        _ast = parse(
            """
            def fn_a(a, b):
                return a + b

            @rattr_results(calls=[("fn_a()", (["alpha", "beta"], {}))])
            def ano():
                return "it is a lie, i do nothing!"
            """
        )
        _ctx = compile_root_context(_ast)

        target = _ctx.get("fn_a")
        fn_def = _ast.body[1]

        expected = {
            "sets": set(),
            "gets": set(),
            "dels": set(),
            "calls": {
                Call(
                    name="fn_a",
                    args=CallArguments(args=("alpha", "beta")),
                    target=target,
                ),
            },
        }

        assert parse_rattr_results_from_annotation(fn_def, context=_ctx) == expected

    def test_is_blacklisted_module(self, stdlib_modules):
        for m in stdlib_modules:
            assert not is_blacklisted_module(m)

        assert is_blacklisted_module("rattr")
        assert is_blacklisted_module("rattr.analyser.error")
        assert is_blacklisted_module("rattr.analyser.a.b.c.d.e")
        assert is_blacklisted_module("rattr.assertors.argument_names")

    def test_is_pip_module(self):
        # NOTE
        #   This is not the best test ever as it assumes that "flask" are
        #   installed and that "numpy" and "pandas" are not.
        #   However, as this test is testing code that if checking for that
        #   exact property I see no better way of testing it -- a mock would
        #   mock away the tested feature.
        assert is_pip_module("isort")
        assert is_pip_module("pytest")

        assert not is_pip_module("numpy")
        assert not is_pip_module("pandas")

        assert not is_pip_module("math")
        assert not is_pip_module("string")

    def test_is_stdlib_module(self, stdlib_modules):
        # Test against known stdlib modules
        for m in stdlib_modules:
            assert is_stdlib_module(m)

        # Test against submodules of stdlib modules
        assert is_stdlib_module("os.path")
        assert is_stdlib_module("os.path.join")
        assert is_stdlib_module("math.sin")
        assert is_stdlib_module("math.pi")

        # Technically false, but true to implementation
        assert is_stdlib_module("math.this.is.not.in.the.stdlib")

        # Test against pip/thirdparty modules
        assert not is_stdlib_module("flask")
        assert not is_stdlib_module("numpy")
        assert not is_stdlib_module("flake8")
        assert not is_stdlib_module("foobar")

        # Test against ratter/firstparty modules
        assert not is_stdlib_module("rattr")
        assert not is_stdlib_module("rattr.models.context")
        assert not is_stdlib_module("rattr.models.context._context")

        assert not is_stdlib_module("anything.dotted.anything")

    def test_is_in_builtins(self, builtins):
        for b in builtins:
            assert is_in_builtins(b)

        assert is_in_builtins("max")
        assert is_in_builtins("reversed")
        assert is_in_builtins("int")
        assert is_in_builtins("enumerate")

        assert not is_in_builtins("math.max")
        assert not is_in_builtins("foo.bar.baz")
        assert not is_in_builtins("foo_bar_baz")
        assert not is_in_builtins("")

    def test_get_function_body(self):
        # Lambda
        fn = ast.parse("fn = lambda: 2 + 3").body[0].value
        assert get_function_body(fn) == [fn.body]

        # Func
        fn = ast.parse("def fn(): pass").body[0]
        assert get_function_body(fn) == fn.body

        # Async
        fn = ast.parse("async def fn(): pass").body[0]
        assert get_function_body(fn) == fn.body

    def test_get_assignment_targets(self):
        # Assign, single
        assign = ast.parse("a = 1").body[0]
        assert get_assignment_targets(assign) == assign.targets

        # Assign, multi
        assign = ast.parse("a, b = 1, 2").body[0]
        assert get_assignment_targets(assign) == assign.targets

        # AnnAssign
        assign = ast.parse("a: int = 1").body[0]
        assert get_assignment_targets(assign) == [assign.target]

        # AugAssign
        assign = ast.parse("a += 1").body[0]
        assert get_assignment_targets(assign) == [assign.target]

    def test_get_assignment_targets_walrus(self):
        # Walrus in assign
        assign = ast.parse("a = (b := 1)").body[0]
        assert get_assignment_targets(assign) == assign.targets

        # Walrus
        assign = ast.parse("a = (b := 1)").body[0].value
        assert get_assignment_targets(assign) == [assign.target]

    def test_get_contained_walruses(self, stringify_nodes, walrus):
        walruses = get_contained_walruses(ast.parse("a = b").body[0])
        assert stringify_nodes(walruses) == []

        walruses = get_contained_walruses(ast.parse("a = (b := c)").body[0])
        assert stringify_nodes(walruses) == stringify_nodes([walrus("b := c")])

        walruses = get_contained_walruses(ast.parse("a = (x := y, m := n)").body[0])
        assert stringify_nodes(walruses) == stringify_nodes(walrus("x := y", "m := n"))

        walruses = get_contained_walruses(ast.parse("a = (b, x := y)").body[0])
        assert stringify_nodes(walruses) == stringify_nodes([walrus("x := y")])

        # Does not nest (node visitor should recurse)
        walruses = get_contained_walruses(ast.parse("a = (b, (c, x := y))").body[0])
        assert stringify_nodes(walruses) == []

    def test_assignment_is_one_to_one(self):
        is_one_to_one = {
            "x = 1",
            "x: int = 1",
            "x += 1",
        }
        is_not_one_to_one = {
            "x, y = z",
            "x, y = a, b",
            "x = y, z",
            "x += y, z",
        }

        for case in is_one_to_one:
            assert assignment_is_one_to_one(ast.parse(case).body[0])

        for case in is_not_one_to_one:
            assert not assignment_is_one_to_one(ast.parse(case).body[0])

    def test_lambda_in_rhs(self):
        # Single
        assign = ast.parse("a = lambda: 1").body[0]
        assert lambda_in_rhs(assign)

        # Multi
        assign = ast.parse("a = 1, lambda: 1").body[0]
        assert lambda_in_rhs(assign)

        # AnnAssign
        assign = ast.parse("a: SomeType = lambda: 1").body[0]
        assert lambda_in_rhs(assign)

        # AugAssign
        assign = ast.parse("a += lambda: 1").body[0]
        assert lambda_in_rhs(assign)

        # Single, not
        assign = ast.parse("a = 1").body[0]
        assert not lambda_in_rhs(assign)

        # Multi, not
        assign = ast.parse("a = 1, 2").body[0]
        assert not lambda_in_rhs(assign)

        # AnnAssign, not
        assign = ast.parse("a: SomeType = SomeType()").body[0]
        assert not lambda_in_rhs(assign)

        # AugAssign, not
        assign = ast.parse("a += 1").body[0]
        assert not lambda_in_rhs(assign)

    def test_walrus_in_rhs(self):
        # Single
        assign = ast.parse("a = (b := c)").body[0]
        assert walrus_in_rhs(assign)

        # Multi
        assign = ast.parse("a = 1, (b := c)").body[0]
        assert walrus_in_rhs(assign)

        # AnnAssign
        assign = ast.parse("a: SomeType = (b := c)").body[0]
        assert walrus_in_rhs(assign)

        # AugAssign
        assign = ast.parse("a += (b := c)").body[0]
        assert walrus_in_rhs(assign)

        # Single, not
        assign = ast.parse("a = 1").body[0]
        assert not walrus_in_rhs(assign)

        # Multi, not
        assign = ast.parse("a = 1, 2").body[0]
        assert not walrus_in_rhs(assign)

        # AnnAssign, not
        assign = ast.parse("a: SomeType = SomeType()").body[0]
        assert not walrus_in_rhs(assign)

        # AugAssign, not
        assign = ast.parse("a += 1").body[0]
        assert not walrus_in_rhs(assign)

    def test_class_in_rhs(self):
        _ctx = Context(None)
        _ctx.add(Name("x"))
        _ctx.add(Name("LooksLikeAClass"))
        _ctx.add(Class("MyClass", interface=CallInterface()))

        # No class in RHS
        assign = ast.parse("a = x").body[0]
        assert not class_in_rhs(assign, _ctx)

        assign = ast.parse("a = LooksLikeAClass()").body[0]
        assert not class_in_rhs(assign, _ctx)

        assign = ast.parse("a, b = x, LooksLikeAClass()").body[0]
        assert not class_in_rhs(assign, _ctx)

        # Class in RHS
        assign = ast.parse("a = MyClass()").body[0]
        assert class_in_rhs(assign, _ctx)

        assign = ast.parse("a, b = x, MyClass()").body[0]
        assert class_in_rhs(assign, _ctx)

    def test_is_starred_import(self):
        # Not starred: normal import
        _import = ast.parse("import math").body[0]
        assert not is_starred_import(_import)

        # Not starred: normal import
        _import = ast.parse("from math import pi").body[0]
        assert not is_starred_import(_import)

        # Not starred: normal import
        _import = ast.parse("from math import cos, sin, tan").body[0]
        assert not is_starred_import(_import)

        # Starred
        _import = ast.parse("from math import *").body[0]
        assert is_starred_import(_import)

        # Starred, relative
        _import = ast.parse("from .here.module import *").body[0]
        assert is_starred_import(_import)

    def test_is_relative_import(self):
        # Not starred: normal import
        _import = ast.parse("import math").body[0]
        assert not is_relative_import(_import)

        # Not starred: normal import
        _import = ast.parse("from math import pi").body[0]
        assert not is_relative_import(_import)

        # Not starred: normal import
        _import = ast.parse("from math import cos, sin, tan").body[0]
        assert not is_relative_import(_import)

        # Starred
        _import = ast.parse("from math import *").body[0]
        assert not is_relative_import(_import)

        # Starred, relative
        _import = ast.parse("from .here.module import *").body[0]
        assert is_relative_import(_import)

    def test_get_function_form(self):
        # No args
        fn = ast.parse("def fn(): pass").body[0]
        assert get_function_form(fn) == "'fn()'"

        # Positional arguments
        fn = ast.parse("def fn(a, b, c): pass").body[0]
        assert get_function_form(fn) == "'fn(a, b, c)'"

        # Complex
        fn = ast.parse("def fn(a, b, c=None, *args, **kwargs): pass").body[0]
        assert get_function_form(fn) == "'fn(a, b, c, *args, **kwargs)'"

    def test_is_excluded_name(self, arguments: ArgumentsFn):
        assert not is_excluded_name("sin")
        assert not is_excluded_name("_hidden_func")

        with arguments(_excluded_names={"sin"}):
            assert is_excluded_name("sin")
            assert not is_excluded_name("_hidden_func")

        with arguments(_excluded_names={"_.*"}):
            assert not is_excluded_name("sin")
            assert is_excluded_name("_hidden_func")

    def test_is_method_on_constant(self, constant: str):
        assert not is_method_on_constant("some_var")
        assert not is_method_on_constant("some_var.methodd")
        assert not is_method_on_constant("some_var[0].method")

        assert not is_method_on_constant(constant)
        assert is_method_on_constant(constant + ".methodd")
        assert is_method_on_constant(constant + ".[0].method")

    def test_is_method_on_cast(self):
        assert not is_method_on_cast("some_var")
        assert not is_method_on_cast("func")
        assert not is_method_on_cast("func")
        assert not is_method_on_cast("func().on_result")

        assert is_method_on_cast("set().union")
        assert is_method_on_cast("list().append")

    def test_is_method_on_primitive(self, constant: str):
        assert not is_method_on_primitive(constant)
        assert not is_method_on_primitive("some_var")
        assert not is_method_on_primitive("func")
        assert not is_method_on_primitive("func")
        assert not is_method_on_primitive("func().on_result")

        assert is_method_on_primitive("set().union")
        assert is_method_on_primitive("list().append")
        assert is_method_on_primitive(constant + ".methodd")
        assert is_method_on_primitive(constant + ".[0].method")

    def test_get_dynamic_name(self):
        call = ast.parse("getattr(object, 'attribute')").body[0].value
        result = get_dynamic_name("getattr", call, "{first}.{second}")
        expected = Name("object.attribute", "object")

        assert result == expected

        call = ast.parse("getattr(getattr(a, 'b'), 'c')").body[0].value
        result = get_dynamic_name("getattr", call, "{first}.{second}")
        expected = Name("a.b.c", "a")

        assert result == expected

        call = ast.parse("get_sub_attr(obj, 'attr')").body[0].value
        result = get_dynamic_name("get_sub_attr", call, "{first}.sub.{second}")
        expected = Name("obj.sub.attr", "obj")

        assert result == expected

        call = ast.parse("get_sub_attr(get_sub_attr(o, 'in'), 'out')").body[0].value
        result = get_dynamic_name("get_sub_attr", call, "{first}.mid.{second}")
        expected = Name("o.in.mid.out", "o")

        assert result == expected

        call = ast.parse("get_sub(obj, ...)").body[0].value
        result = get_dynamic_name("get_sub", call, "{first}.sub")
        expected = Name("obj.sub", "obj")

        assert result == expected

        call = ast.parse("getattr(another_function(a, 'b'), 'c')").body[0].value
        with mock.patch("sys.exit") as _exit:
            result = get_dynamic_name("getattr", call, "{first}.{second}")

        assert _exit.call_count == 1

    def test_get_attrname(self):
        attr = ast.parse("object.attribute").body[0].value
        assert get_attrname(attr) == "attribute"

        attr = ast.parse("a.b.c.d").body[0].value
        assert get_attrname(attr) == "d"

        attr = ast.parse("just_some_var").body[0].value
        assert get_attrname(attr) == "just_some_var"

        attr = ast.parse("a(arg, arg, arg).b[0].c").body[0].value
        assert get_attrname(attr) == "c"

        attr = ast.parse("def fn(): pass").body[0]
        with pytest.raises(TypeError):
            assert get_attrname(attr)

    def test_dict_changes(self):
        ir = {"a": 1, "b": 2, "c": 3, "d": 4}

        # No changes
        with DictChanges(ir) as diff:
            pass

        assert not diff.added
        assert not diff.removed

        # Changes
        with DictChanges(ir) as diff:
            ir["x"] = 99
            del ir["a"]
            del ir["b"]

        assert diff.added == {"x"}
        assert diff.removed == {"a", "b"}


class TestNamedTupleInRhs:
    def test_namedtuple_in_rhs(self):
        ast_ = ast.parse("point = namedtuple('point', ['x', 'y'])")
        assignment = ast_.body[0]
        assert namedtuple_in_rhs(assignment)

    def test_user_namedtuple_in_rhs(self):
        ast_ = ast.parse("point = user_extended.namedtuple('point', ['x', 'y'])")
        assignment = ast_.body[0]
        assert namedtuple_in_rhs(assignment)

    def test_namedtuple_not_in_rhs(self):
        ast_ = ast.parse("point = noomedtoople('point', ['x', 'y'])")
        assignment = ast_.body[0]
        assert not namedtuple_in_rhs(assignment)


class TestGetNamedTupleAttrsFromCall:
    def test_rhs_is_not_call(self):
        ast_ = ast.parse("point = 'not_a_call'")
        assignment = ast_.body[0]

        with pytest.raises(TypeError):
            get_namedtuple_attrs_from_call(assignment)

    def test_too_few_arguments(self):
        ast_ = ast.parse("point = namedtuple('point')")
        assignment = ast_.body[0]

        with pytest.raises(SystemExit):
            get_namedtuple_attrs_from_call(assignment)

    def test_too_many_arguments(self):
        ast_ = ast.parse("point = namedtuple('point', 'x', 'y')")
        assignment = ast_.body[0]

        with pytest.raises(SystemExit):
            get_namedtuple_attrs_from_call(assignment)

    def test_second_argument_is_not_a_list(self):
        ast_ = ast.parse("point = namedtuple('point', ('x', 'y'))")
        assignment = ast_.body[0]

        with pytest.raises(SystemExit):
            get_namedtuple_attrs_from_call(assignment)

    def test_second_argument_is_not_a_list_of_string_literals(self):
        # Non-string literal
        ast_ = ast.parse("point = namedtuple('point', ['x', 123])")
        assignment = ast_.body[0]

        with pytest.raises(SystemExit):
            get_namedtuple_attrs_from_call(assignment)

        # Non-literal
        ast_ = ast.parse("point = namedtuple('point', ['x', some_variable])")
        assignment = ast_.body[0]

        with pytest.raises(SystemExit):
            get_namedtuple_attrs_from_call(assignment)

    def test_attrs_from_named_tuple_call_the_empty_list(self):
        # Non-string literal
        ast_ = ast.parse("point = namedtuple('point', [])")
        assignment = ast_.body[0]

        assert get_namedtuple_attrs_from_call(assignment) == ["self"]

    def test_attrs_from_named_tuple_call_one_item(self):
        # Non-string literal
        ast_ = ast.parse("point = namedtuple('point', ['x'])")
        assignment = ast_.body[0]

        assert get_namedtuple_attrs_from_call(assignment) == ["self", "x"]

    def test_attrs_from_named_tuple_call_two_items(self):
        # Non-string literal
        ast_ = ast.parse("point = namedtuple('point', ['x', 'y'])")
        assignment = ast_.body[0]

        assert get_namedtuple_attrs_from_call(assignment) == ["self", "x", "y"]

    def test_attrs_from_named_tuple_call_multiple_items(self):
        # Non-string literal
        ast_ = ast.parse("vec4 = vec4('vector', ['a', 'b', 'c', 'd'])")
        assignment = ast_.body[0]

        expected = ["self", "a", "b", "c", "d"]
        assert get_namedtuple_attrs_from_call(assignment) == expected



from __future__ import annotations

from typing import TYPE_CHECKING

from rattr.analyser.file import FileAnalyser
from rattr.models.context import Context, compile_root_context
from rattr.models.ir import FileIr
from rattr.models.symbol import (
    Builtin,
    Call,
    CallArguments,
    CallInterface,
    Func,
    Import,
    Name,
)

if TYPE_CHECKING:
    from tests.shared import MakeSymbolTableFn, ParseFn


class TestCustomCollectionsAnalysers:
    def test_defaultdict_no_factory(
        self,
        parse: ParseFn,
        make_symbol_table: MakeSymbolTableFn,
    ):
        _ast = parse(
            """
            from collections import defaultdict

            def a_func(arg):
                d = defaultdict()
            """
        )
        results = FileAnalyser(_ast, compile_root_context(_ast)).analyse()

        defaultdict_import = Import("defaultdict", "collections.defaultdict")
        a_func = Func(name="a_func", interface=CallInterface(args=("arg",)))

        expected = FileIr(
            context=Context(
                parent=None,
                symbol_table=make_symbol_table(
                    [defaultdict_import, a_func],
                    include_root_symbols=True,
                ),
            ),
            file_ir={
                a_func: {
                    "gets": set(),
                    "sets": {Name("d")},
                    "dels": set(),
                    "calls": set(),
                }
            },
        )

        assert results == expected

    def test_defaultdict_named_factory(
        self,
        parse: ParseFn,
        make_symbol_table: MakeSymbolTableFn,
    ):
        _ast = parse(
            """
            from collections import defaultdict

            def a_func(arg):
                d = defaultdict(factory)

            def factory():
                return hello.results
            """
        )
        results = FileAnalyser(_ast, compile_root_context(_ast)).analyse()

        defaultdict_import = Import("defaultdict", "collections.defaultdict")
        a_func = Func(name="a_func", interface=CallInterface(args=("arg",)))
        factory = Func(name="factory", interface=CallInterface(args=()))

        expected = FileIr(
            context=Context(
                parent=None,
                symbol_table=make_symbol_table(
                    [
                        defaultdict_import,
                        a_func,
                        factory,
                    ],
                    include_root_symbols=True,
                ),
            ),
            file_ir={
                a_func: {
                    "gets": set(),
                    "sets": {
                        Name("d"),
                    },
                    "dels": set(),
                    "calls": {Call("factory", args=CallArguments(), target=factory)},
                },
                factory: {
                    "gets": {
                        Name("hello.results", "hello"),
                    },
                    "sets": set(),
                    "dels": set(),
                    "calls": set(),
                },
            },
        )

        assert results == expected

    def test_defaultdict_lambda_factory(
        self,
        parse: ParseFn,
        make_symbol_table: MakeSymbolTableFn,
    ):
        defaultdict_import = Import("defaultdict", "collections.defaultdict")
        a_func = Func(name="a_func", interface=CallInterface(args=("arg",)))

        # Lambda to literal
        _ast = parse(
            """
            from collections import defaultdict

            def a_func(arg):
                d = defaultdict(lambda: 0)
            """
        )
        results = FileAnalyser(_ast, compile_root_context(_ast)).analyse()

        expected = FileIr(
            context=Context(
                parent=None,
                symbol_table=make_symbol_table(
                    [
                        defaultdict_import,
                        a_func,
                    ],
                    include_root_symbols=True,
                ),
            ),
            file_ir={
                a_func: {
                    "gets": set(),
                    "sets": {
                        Name("d"),
                    },
                    "dels": set(),
                    "calls": set(),
                }
            },
        )

        assert results == expected

        # Lambda to attr
        _ast = parse(
            """
            from collections import defaultdict

            def a_func(arg):
                d = defaultdict(lambda: arg.attr)
            """
        )
        results = FileAnalyser(_ast, compile_root_context(_ast)).analyse()

        expected = FileIr(
            context=Context(
                parent=None,
                symbol_table=make_symbol_table(
                    [
                        defaultdict_import,
                        a_func,
                    ],
                    include_root_symbols=True,
                ),
            ),
            file_ir={
                a_func: {
                    "gets": {Name("arg.attr", "arg")},
                    "sets": {
                        Name("d"),
                    },
                    "dels": set(),
                    "calls": set(),
                }
            },
        )

        assert results == expected

    def test_defaultdict_nested_factory(
        self,
        parse: ParseFn,
        make_symbol_table: MakeSymbolTableFn,
    ):
        _ast = parse(
            """
            from collections import defaultdict

            def a_func(arg):
                d = defaultdict(defaultdict(list))
            """
        )
        results = FileAnalyser(_ast, compile_root_context(_ast)).analyse()

        defaultdict_import = Import("defaultdict", "collections.defaultdict")
        a_func = Func(name="a_func", interface=CallInterface(args=("arg",)))
        list_builtin = Builtin("list")

        expected = FileIr(
            context=Context(
                parent=None,
                symbol_table=make_symbol_table(
                    [
                        defaultdict_import,
                        a_func,
                    ],
                    include_root_symbols=True,
                ),
            ),
            file_ir={
                a_func: {
                    "gets": set(),
                    "sets": {
                        Name("d"),
                    },
                    "dels": set(),
                    "calls": {Call("list", args=CallArguments(), target=list_builtin)},
                }
            },
        )

        assert results == expected



from __future__ import annotations

import ast
import re

import pytest

from rattr import error
from rattr.ast.util import (
    NAMEDTUPLE_INVALID_SECOND_PARAMETER_VALUE_ERROR,
    NAMEDTUPLE_INVALID_SIGNATURE_ERROR,
    assignment_is_one_to_one,
    assignment_targets,
    basename_of,
    fullname_of,
    get_python_attr_access_fn_obj_attr_pair,
    has_lambda_in_rhs,
    has_namedtuple_declaration_in_rhs,
    has_walrus_in_rhs,
    is_call_to_fn,
    is_relative_import,
    is_starred_import,
    is_string_literal,
    namedtuple_init_signature_from_declaration,
    names_of,
    parse_space_delimited_ast_string,
    unpack_ast_list_of_strings,
    unravel_names,
    walruses_in_rhs,
)
from rattr.models.symbol._symbols import PYTHON_ATTR_ACCESS_BUILTINS

# TODO Migrate tests from old utils functions


class TestIsCallToFn:
    @pytest.mark.parametrize("expr", ["thing = ..."])
    def test_not_a_call(self, expr):
        with pytest.raises(TypeError):
            is_call_to_fn(ast.parse(expr).body[0].value, "blah")

    @pytest.mark.parametrize(
        "expr,target",
        [
            ("fn(a, b, c)", "fn"),
            ("outer(inner(a, b, c))", "outer"),
        ],
    )
    def test_positive_case(self, expr, target):
        assert is_call_to_fn(ast.parse(expr).body[0].value, target)

    @pytest.mark.parametrize(
        "expr,target",
        [
            ("fn(a, b, c)", "not_fn"),
            ("outer(inner(a, b, c))", "inner"),
            ("outer(inner(a, b, c))", "neither_inner_nor_outer"),
            ("obj.method(a, b, c)", "different_func"),
            ("obj.method(a, b, c)", "obj"),
            ("obj.method(a, b, c)", "method"),
            ("obj.method(a, b, c)", "obj.method"),
            ("fn(a, b, c).on_result()", "not_fn"),
            ("fn(a, b, c).on_result()", "on_result"),
            ("fn(a, b, c).on_result()", "not_fn.on_result"),
            ("fn(a, b, c).on_result()", "not_fn().on_result"),
        ],
    )
    def test_negative_case(self, expr, target):
        assert not is_call_to_fn(ast.parse(expr).body[0].value, target)

    def test_nested_case(self):
        call: ast.Call = ast.parse("fn(a, b, c).on_result()").body[0].value

        assert not is_call_to_fn(call, "fn")
        assert is_call_to_fn(call.func.value, "fn")


class TestIsStringLiteral:
    @pytest.mark.parametrize("literal", ["'some_string'", "'1234'", "'True'"])
    def test_constant_positive_case(self, literal):
        assert is_string_literal(ast.parse(literal).body[0].value)

    @pytest.mark.parametrize("literal", ["some_var", "1234", "True"])
    def test_constant_negative_case(self, literal):
        assert not is_string_literal(ast.parse(literal).body[0].value)

    @pytest.mark.parametrize("expr", ["x = 5", "y = 'string'"])
    def test_negative_case(self, expr):
        assert not is_string_literal(ast.parse(expr).body[0])


class TestGetPythonAttrAccessFnObjAttrPair:
    @pytest.mark.parametrize(
        "expr,result",
        [
            ("getattr(a, 'b')", ("a", "b")),
            ("getattr(getattr(a, 'b'), 'c')", ("a.b", "c")),
            ("getattr(getattr(a.b[0], 'c'), 'd')", ("a.b[].c", "d")),
        ],
    )
    def test_docstring_examples(self, expr, result):
        assert (
            get_python_attr_access_fn_obj_attr_pair(
                "getattr", ast.parse(expr).body[0].value
            )
            == result
        )

    @pytest.mark.parametrize(
        "expr,result",
        [
            ("getattr(a, some_string_variable)", ("a", "<some_string_variable>")),
            ("getattr(getattr(a, var), 'attr')", ("a.<var>", "attr")),
        ],
    )
    def test_docstring_error_examples(self, arguments, expr, result):
        _expr = ast.parse(expr).body[0].value

        with arguments(is_strict=True):
            with pytest.raises(SystemExit):
                get_python_attr_access_fn_obj_attr_pair("getattr", _expr, warn=True)

        assert (
            get_python_attr_access_fn_obj_attr_pair("getattr", _expr, warn=True)
            == result
        )

    @pytest.mark.parametrize(
        "expr",
        [
            ("getattr(a or b, 'c')"),
            ("getattr(getattr(a1 or a2, 'b'), 'c')"),
        ],
    )
    def test_expr_is_illegal(self, expr):
        with pytest.raises(TypeError):
            get_python_attr_access_fn_obj_attr_pair(
                "getattr",
                ast.parse(expr).body[0].value,
            )

    def test_illegal_nesting(self):
        _expr = ast.parse("hasattr(getattr(obj, 'thing'), 'attr')").body[0].value

        with pytest.raises(SystemExit):
            get_python_attr_access_fn_obj_attr_pair("hasattr", _expr)

    @pytest.mark.parametrize("fn", PYTHON_ATTR_ACCESS_BUILTINS)
    def test_support_all_attr_accessors(self, fn):
        _expr = ast.parse(f"{fn}(obj, 'attr')").body[0].value
        assert get_python_attr_access_fn_obj_attr_pair(fn, _expr) == ("obj", "attr")


class TestNameOf:
    @pytest.mark.parametrize(
        "expr,basename,fullname",
        [
            ("my_variable", "my_variable", "my_variable"),
            ("my_variable.attr", "my_variable", "my_variable.attr"),
            ("my_variable[100].attr", "my_variable", "my_variable[].attr"),
            ("*my_variable[100].attr", "my_variable", "*my_variable[].attr"),
            ("(a, b)[0].attr", "@Tuple", "@Tuple[].attr"),
        ],
    )
    def test_docstring_examples(self, expr, basename, fullname):
        assert names_of(ast.parse(expr).body[0].value, safe=True) == (
            basename,
            fullname,
        )

    @pytest.mark.parametrize(
        "expr,error",
        [
            ("(a, b)[0].attr", error.RattrLiteralInNameable),
        ],
    )
    def test_docstring_error_examples(self, expr, error):
        with pytest.raises(error):
            names_of(ast.parse(expr).body[0].value, safe=False)


class TestBasenameOf:
    # See TestNameOf

    @pytest.mark.parametrize(
        "expr, basename",
        [
            ("var", "var"),
            ("var.attr", "var"),
        ],
    )
    def test_basename_od(self, expr, basename):
        assert basename_of(ast.parse(expr).body[0].value) == basename


class TestFullnameOf:
    # See TestNameOf

    @pytest.mark.parametrize(
        "expr, basename",
        [
            ("var", "var"),
            ("var.attr", "var.attr"),
        ],
    )
    def test_fullname_od(self, expr, basename):
        assert fullname_of(ast.parse(expr).body[0].value) == basename


class TestUnravelNames:
    @pytest.mark.parametrize(
        "expr,expected",
        [
            ("a = 1", ["a"]),
            ("a.my_attr = 1", ["a"]),
            ("a, b = 1, 2", ["a", "b"]),
            ("(a, b), c = [1, 2], 3", ["a", "b", "c"]),
            ("(a, b), c, d.e = 1, 2, 3, 4", ["a", "b", "c", "d"]),
        ],
    )
    def test_unravel_via_basename(self, expr, expected):
        names = ast.parse(expr).body[0].targets[0]
        assert unravel_names(names) == expected

    @pytest.mark.parametrize(
        "expr,expected",
        [
            ("a = 1", ["a"]),
            ("a.my_attr = 1", ["a.my_attr"]),
            ("a, b = 1, 2", ["a", "b"]),
            ("(a, b), c = [1, 2], 3", ["a", "b", "c"]),
            ("(a, b), c, d.e = 1, 2, 3, 4", ["a", "b", "c", "d.e"]),
        ],
    )
    def test_unravel_via_fullname(self, expr, expected):
        names = ast.parse(expr).body[0].targets[0]
        assert unravel_names(names, _get_name=fullname_of) == expected


class TestIsStarredImport:
    @pytest.mark.parametrize(
        "stmt",
        [
            "from math import *",
            "from .relative.module import *",
        ],
    )
    def test_is_starred_import(self, stmt):
        assert is_starred_import(ast.parse(stmt).body[0])

    @pytest.mark.parametrize(
        "stmt",
        [
            "import math",
            "from math import pi",
            "from math import sin, cos, tan",
        ],
    )
    def test_is_not_starred_import(self, stmt):
        assert not is_starred_import(ast.parse(stmt).body[0])


class TestIsRelativeImport:
    @pytest.mark.parametrize(
        "stmt",
        [
            "from .relative.module import my_util",
            "from .relative.module import MyClass, my_util, my_other_util",
            "from .relative.module import *",
        ],
    )
    def test_is_relative_import(self, stmt):
        assert is_relative_import(ast.parse(stmt).body[0])

    @pytest.mark.parametrize(
        "stmt",
        [
            "import math",
            "from math import pi",
            "from math import sin, cos, tan",
            "from math import *",
        ],
    )
    def test_is_not_relative_import(self, stmt):
        assert not is_relative_import(ast.parse(stmt).body[0])


@pytest.mark.parametrize(
    "expr, expected",
    testcases := [
        ("a = 1", ["a"]),
        ("a, b = 1, 2", ["a", "b"]),
        ("a: int = 1", ["a"]),
        ("a += 1", ["a"]),
    ],
    ids=[t[0] for t in testcases],
)
def test_assignment_targets(expr: str, expected: list[ast.expr]):
    # We must convert to str and compare as `ast.Name(id="a") != ast.Name(id="a")` as
    # ast node equality is an `is` check.
    targets = [
        name
        for target in assignment_targets(ast.parse(expr).body[0])
        for name in unravel_names(target)
    ]
    assert targets == expected


def test_assignment_targets_in_walrus():
    parsed = ast.parse("a = (b := 1)")

    a: ast.Assign = parsed.body[0]
    b: ast.NamedExpr = parsed.body[0].value

    assert assignment_targets(a) == a.targets
    assert assignment_targets(b) == [b.target]


@pytest.mark.parametrize(
    "expr, expected",
    testcases := [
        ("x = 1", True),
        ("x: int = 1", True),
        ("x += 1", True),
        ("x, y = z", False),
        ("x, y = a, b", False),
        ("x = y, z", False),
        ("x += y, z", False),
    ],
    ids=[t[0] for t in testcases],
)
def test_assignment_is_one_to_one(expr: str, expected: bool):
    assert assignment_is_one_to_one(ast.parse(expr).body[0]) == expected


@pytest.mark.parametrize(
    "expr, expected",
    testcases := [
        ("a = lambda: 1", True),
        ("a = 1, lambda: 1", True),
        ("a: SomeType = lambda: 1", True),
        ("a += lambda: 1", True),
        ("a = 1", False),
        ("a = 1, 2", False),
        ("a: SomeType = SomeType()", False),
        ("a += 1", False),
    ],
    ids=[t[0] for t in testcases],
)
def test_has_lambda_in_rhs(expr: str, expected: bool):
    assert has_lambda_in_rhs(ast.parse(expr).body[0]) == expected


@pytest.mark.parametrize(
    "expr, expected",
    testcases := [
        ("a = (b := c)", True),
        ("a = 1, (b := c)", True),
        ("a: SomeType = (b := c)", True),
        ("a += (b := c)", True),
        ("a = 1", False),
        ("a = 1, 2", False),
        ("a: SomeType = SomeType()", False),
        ("a += 1", False),
    ],
    ids=[t[0] for t in testcases],
)
def test_has_walrus_in_rhs(expr: str, expected: bool):
    assert has_walrus_in_rhs(ast.parse(expr).body[0]) == expected


@pytest.mark.parametrize(
    "expr, expected",
    testcases := [
        ("a = b", []),
        ("a = (b := c)", ["b := c"]),
        ("a = (x := y, m := n)", ["x := y", "m := n"]),
        ("a = (b, x := y)", ["x := y"]),
        ("a = (b, (c, x := y))", []),  # we don't handle nested walruses
    ],
    ids=[t[0] for t in testcases],
)
def test_walruses_in_rhs(walrus, expr: str, expected: list[str]):
    # We stringify the nodes here as the lhs and rhs will be equivalent but have
    # different addresses so fail "==" (which is an "is" check for ast.NamedExpr it
    # seems)
    actual_as_string = [ast.dump(e) for e in walruses_in_rhs(ast.parse(expr).body[0])]
    expected_as_string = [ast.dump(walrus(e)) for e in expected]
    assert actual_as_string == expected_as_string


@pytest.mark.parametrize(
    "expr, expected",
    testcases := [
        ("point = namedtuple('point', ['x', 'y'])", True),
        ("point = user_extended.namedtuple('point', ['x', 'y'])", True),
        ("point = noomedtoople('point', ['x', 'y'])", False),
    ],
    ids=[t[0] for t in testcases],
)
def test_has_namedtuple_declaration_in_rhs(expr: str, expected: bool):
    assert has_namedtuple_declaration_in_rhs(ast.parse(expr).body[0]) == expected


def test_namedtuple_init_signature_from_declaration_not_a_call():
    with pytest.raises(TypeError):
        namedtuple_init_signature_from_declaration(ast.parse("'not_a_call'").body[0])


@pytest.mark.parametrize(
    "expr, error",
    testcases := [
        (
            "namedtuple('point')",
            re.escape(NAMEDTUPLE_INVALID_SIGNATURE_ERROR),
        ),
        (
            "namedtuple('point', 'x', 'y')",
            re.escape(NAMEDTUPLE_INVALID_SIGNATURE_ERROR),
        ),
        (
            "namedtuple('point', ('x', 'y'))",
            re.escape(NAMEDTUPLE_INVALID_SECOND_PARAMETER_VALUE_ERROR),
        ),
        (
            "namedtuple('point', ['x', 123])",
            re.escape(NAMEDTUPLE_INVALID_SECOND_PARAMETER_VALUE_ERROR),
        ),
        (
            "namedtuple('point', ['x', some_variable])",
            re.escape(NAMEDTUPLE_INVALID_SECOND_PARAMETER_VALUE_ERROR),
        ),
    ],
    ids=[t[0] for t in testcases],
)
def test_namedtuple_init_signature_from_declaration_invalid(expr: str, error: str):
    with pytest.raises(ValueError, match=error):
        namedtuple_init_signature_from_declaration(ast.parse(expr).body[0])


@pytest.mark.parametrize(
    "expr, attributes",
    testcases := [
        ("namedtuple('point', [])", ["self"]),
        ("namedtuple('point', ['x'])", ["self", "x"]),
        ("namedtuple('point', ['x', 'y'])", ["self", "x", "y"]),
        ("vec4('vector', ['a', 'b', 'c', 'd'])", ["self", "a", "b", "c", "d"]),
    ],
    ids=[t[0] for t in testcases],
)
def test_namedtuple_init_signature_from_declaration(expr: str, attributes: list[str]):
    assert (
        namedtuple_init_signature_from_declaration(ast.parse(expr).body[0])
        == attributes
    )


@pytest.mark.parametrize(
    "expr, expected",
    testcases := [
        ("[]", []),
        ("['a', 'b', 'c']", ["a", "b", "c"]),
    ],
    ids=[t[0] for t in testcases],
)
def test_unpack_ast_list_of_strings(expr: str, expected: list[str]):
    assert unpack_ast_list_of_strings(ast.parse(expr).body[0].value) == expected


def test_unpack_ast_list_of_strings_invalid():
    with pytest.raises(SyntaxError):
        unpack_ast_list_of_strings(ast.parse("[not_a_string_literal]").body[0].value)


@pytest.mark.parametrize(
    "expr, expected",
    testcases := [
        ("''", []),
        ("'a b c'", ["a", "b", "c"]),
    ],
    ids=[t[0] for t in testcases],
)
def test_parse_space_delimited_ast_string(expr: str, expected: list[str]):
    assert parse_space_delimited_ast_string(ast.parse(expr).body[0].value) == expected


def test_parse_space_delimited_ast_string_invalid():
    with pytest.raises(SyntaxError):
        parse_space_delimited_ast_string(ast.parse("blah 0invalid").body[0].value)



from __future__ import annotations

import ast
import builtins
from pathlib import Path
from typing import TYPE_CHECKING, Union

import attrs
from attrs import field

from rattr.codegen import gen_import_from_stmt
from rattr.models.symbol._symbol import (
    AnyCallInterface,
    CallArguments,
    CallInterface,
    Location,
    Symbol,
)
from rattr.models.symbol.util import (
    PYTHON_BUILTINS_LOCATION,
    get_basename_from_name,
    without_call_brackets,
)
from rattr.module_locator.models import ModuleSpec
from rattr.module_locator.util import find_module_name_and_spec

if TYPE_CHECKING:
    from typing import Final, Literal

    from rattr.ast.types import Identifier
    from rattr.module_locator.util import ModuleName


PYTHON_LITERAL_BUILTINS: Final = ("None", "True", "False", "Ellipsis")
"""Python's literal builtin's by identifier."""

PYTHON_ATTR_ACCESS_BUILTINS: Final = ("delattr", "getattr", "hasattr", "setattr")
"""Python's builtin functions to access attributes, i.e. `getattr` etc."""

PYTHON_BUILTINS: Final = tuple(
    b
    for b in dir(builtins)
    if b not in PYTHON_LITERAL_BUILTINS
    if not b.startswith("__")
)
"""Python's callable builtins."""

PYTHON_NON_PRIMITIVE_RETURNING_BUILTINS: Final = frozenset(
    (
        "eval",
        "exec",
        "getattr",
        "iter",
        "next",
        "slice",
        "super",
        "type",
    )
)
"""Python's builtins may-or-will return a non-primitive."""


@attrs.frozen
class Name(Symbol):
    name: str = field()
    basename: str = field()

    token: Union[ast.AST, None] = field(
        default=None,
        kw_only=True,
        hash=False,
        eq=False,
    )
    location: Location = field(kw_only=True, hash=False, eq=False)

    interface: Union[CallInterface, None] = field(default=None, kw_only=True)

    @basename.default
    def _basename_default(self) -> str:
        return get_basename_from_name(self.name)

    @location.default
    def _location_default(self) -> Location:
        if self.token is None:
            return Location(lineno=1, col_offset=0)
        return Location.from_ast_token(self.token)


@attrs.frozen
class Builtin(Symbol):
    name: str = field()

    token: Union[ast.AST, None] = field(
        default=None,
        kw_only=True,
        hash=False,
        eq=False,
    )
    location: Location = field(kw_only=True, hash=False, eq=False)

    interface: AnyCallInterface = field(factory=AnyCallInterface, kw_only=True)

    @location.default
    def _location_default(self) -> Location:
        file = Path(PYTHON_BUILTINS_LOCATION)

        if self.token is None:
            return Location(lineno=1, col_offset=0, file=file)

        return Location.from_ast_token(self.token, file=file)

    @property
    def has_affect(self) -> bool:
        return self.name in PYTHON_ATTR_ACCESS_BUILTINS


@attrs.frozen
class Import(Symbol):
    name: str = field()
    qualified_name: str = field()

    token: Union[ast.AST, None] = field(
        default=None,
        kw_only=True,
        hash=False,
        eq=False,
    )
    location: Location = field(kw_only=True, hash=False, eq=False)

    interface: AnyCallInterface = field(factory=AnyCallInterface, kw_only=True)

    @qualified_name.default
    def _qualified_name_default(self) -> str:
        return self.name

    @location.default
    def _location_default(self) -> Location:
        if self.token is None:
            return Location(lineno=1, col_offset=0)
        return Location.from_ast_token(self.token)

    @property
    def id(self) -> str:
        # We must prepend the qualified name to starred-imports or else they'd all have
        # the same name ("*")!
        if self.name == "*":
            return f"{self.qualified_name}.*"
        return self.name

    @property
    def _module_name_and_spec(
        self,
    ) -> tuple[ModuleName, ModuleSpec] | tuple[None, None]:
        return find_module_name_and_spec(self.qualified_name)

    @property
    def module_name(self) -> ModuleName | None:
        return self._module_name_and_spec[0]

    @property
    def module_spec(self) -> ModuleSpec | None:
        return self._module_name_and_spec[1]

    @property
    def is_import(self) -> Literal[True]:
        return True

    @property
    def origin(self) -> Path | None:
        if self.module_spec is None:
            return None

        if self.module_spec.origin is None:
            return None

        return Path(self.module_spec.origin).resolve()

    def code(self) -> str:
        # NOTE Derive from `self.qualified_name` as `self.module_name` can be `None`
        module = self.qualified_name.replace(f".{self.name}", "")
        return gen_import_from_stmt(module, self.name)


@attrs.frozen
class Func(Symbol):
    name: str = field(converter=without_call_brackets)

    token: Union[ast.AST, None] = field(
        default=None,
        kw_only=True,
        hash=False,
        eq=False,
    )
    location: Location = field(kw_only=True, hash=False, eq=False)

    interface: CallInterface = field(kw_only=True)

    is_async: bool = field(default=False, kw_only=True)

    @location.default
    def _location_default(self) -> Location:
        if self.token is None:
            return Location(lineno=1, col_offset=0)
        return Location.from_ast_token(self.token)

    @classmethod
    def from_fn_def(
        cls: type[Func],
        fn: ast.FunctionDef | ast.AsyncFunctionDef,
    ) -> Func:
        """Return a new `Func` parsed from the given function def."""
        return Func(
            name=fn.name,
            token=fn,
            interface=CallInterface.from_fn_def(fn),
            is_async=isinstance(fn, ast.AsyncFunctionDef),
        )


@attrs.frozen
class Class(Symbol):
    name: str = field(converter=without_call_brackets)

    token: Union[ast.AST, None] = field(
        default=None,
        kw_only=True,
        hash=False,
        eq=False,
    )
    location: Location = field(kw_only=True, hash=False, eq=False)

    interface: CallInterface = field(factory=AnyCallInterface, kw_only=True)

    @location.default
    def _location_default(self) -> Location:
        if self.token is None:
            return Location(lineno=1, col_offset=0)
        return Location.from_ast_token(self.token)

    def with_init(self, init: ast.FunctionDef) -> Class:
        """Return a copy of the class with the initialiser set to the given function."""
        return attrs.evolve(self, interface=CallInterface.from_fn_def(init))

    def with_init_arguments(
        self,
        *,
        posonlyargs: tuple[str, ...] = (),
        args: tuple[str, ...] = (),
        vararg: str | None = None,
        kwonlyargs: tuple[str, ...] = (),
        kwarg: str | None = None,
    ) -> Class:
        return attrs.evolve(
            self,
            interface=CallInterface(
                posonlyargs=posonlyargs,
                args=args,
                vararg=vararg,
                kwonlyargs=kwonlyargs,
                kwarg=kwarg,
            ),
        )

    def with_init_interface(self, interface: CallInterface) -> Class:
        return attrs.evolve(self, interface=interface)

    @classmethod
    def from_class_def(cls: type[Class], ast_class: ast.ClassDef) -> None:
        """Return a new `Class` parsed from the given class def."""
        init_interface = AnyCallInterface()

        for stmt in ast_class.body:
            if not isinstance(stmt, (ast.FunctionDef, ast.AsyncFunctionDef)):
                continue

            if stmt.name != "__init__":
                continue

            init_interface = CallInterface.from_fn_def(stmt)

        return Class(name=ast_class.name, token=ast_class, interface=init_interface)


@attrs.frozen
class Call(Symbol):
    name: str = field(converter=without_call_brackets)

    args: CallArguments = field(factory=CallArguments)
    target: Union[Builtin, Import, Func, Class, Name, None] = field(default=None)

    token: Union[ast.AST, None] = field(
        default=None,
        kw_only=True,
        hash=False,
        eq=False,
    )
    location: Location = field(kw_only=True, hash=False, eq=False)

    interface: Union[CallInterface, None] = field(
        init=False,
        default=None,
        kw_only=True,
    )

    @location.default
    def _location_default(self) -> Location:
        if self.token is None:
            return Location(lineno=1, col_offset=0)
        return Location.from_ast_token(self.token)

    @classmethod
    def from_call(
        cls: type[Call],
        name: str,
        call: ast.Call,
        target: Builtin | Import | Func | Class | None,
        *,
        self: Identifier | None = None,
    ) -> Call:
        return Call(
            name=name,
            args=CallArguments.from_call(call, self=self),
            target=target,
            token=call,
        )

    @property
    def name_of_call(self) -> str:
        return f"{self.name}()"



from __future__ import annotations

from rattr.error.error import (
    error,
    fatal,
    get_file_and_line_info,
    info,
    rattr,
    warning,
)
from rattr.error.exc import (
    RattrBinOpInNameable,
    RattrComprehensionInNameable,
    RattrConstantInNameable,
    RattrLiteralInNameable,
    RattrUnaryOpInNameable,
    RattrUnsupportedError,
)

__all__ = [
    "RattrBinOpInNameable",
    "RattrComprehensionInNameable",
    "RattrConstantInNameable",
    "RattrLiteralInNameable",
    "RattrUnaryOpInNameable",
    "RattrUnsupportedError",
    "error",
    "fatal",
    "get_file_and_line_info",
    "info",
    "rattr",
    "warning",
]



"""Run the full analyser (in single file mode) on a snippet of code.

This should test source code snippets containing the use of a single feature, to ensure
that when run through the full analyser the end results and IR are as expected.
"""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

import pytest

import tests.helpers as helpers
from rattr.models.ir import FileIr
from rattr.models.results import FileResults
from rattr.models.symbol import (
    Builtin,
    Call,
    CallArguments,
    CallInterface,
    Func,
    Name,
)

if TYPE_CHECKING:
    from collections.abc import Callable, Iterator

    from tests.shared import ArgumentsFn, MakeRootContextFn, StateFn


@pytest.fixture(autouse=True)
def __set_current_file(state: StateFn) -> Iterator[None]:
    with state(current_file=Path("my_test_file.py")):
        yield


@pytest.fixture(autouse=True)
def __set_warning_level(arguments: ArgumentsFn) -> Iterator[None]:
    with arguments(_warning_level="all"):
        yield


class TestGeneratorExpession:
    def test_simple_generator_expression(
        self,
        analyse_single_file: Callable[[str], tuple[FileIr, FileResults]],
        make_root_context: MakeRootContextFn,
        capfd: pytest.CaptureFixture[str],
    ):
        file_ir, file_results = analyse_single_file(
            """
            def sum_attrs(xs):
                return sum(x.attr for x in xs)
            """
        )

        _, stderr = capfd.readouterr()
        assert helpers.stdout_matches(stderr, [])

        sum_attrs_symbol = Func("sum_attrs", interface=CallInterface(args=("xs",)))
        expected_file_ir = FileIr(
            context=make_root_context([sum_attrs_symbol], include_root_symbols=True),
            file_ir={
                sum_attrs_symbol: {
                    "gets": {
                        Name("xs"),
                        Name("x.attr", "x"),
                    },
                    "sets": {
                        Name("x"),
                    },
                    "dels": set(),
                    "calls": {
                        Call(
                            name="sum",
                            args=CallArguments(args=("@GeneratorExp",)),
                            target=Builtin(name="sum"),
                        )
                    },
                }
            },
        )
        expected_file_results = FileResults(
            {
                sum_attrs_symbol.id: {
                    "gets": {"xs", "x.attr"},
                    "sets": {"x"},
                    "dels": set(),
                    "calls": {"sum()"},
                }
            }
        )

        assert file_ir == expected_file_ir
        assert file_results == expected_file_results

    def test_generator_expression_with_ifs(
        self,
        analyse_single_file: Callable[[str], tuple[FileIr, FileResults]],
        make_root_context: MakeRootContextFn,
        capfd: pytest.CaptureFixture[str],
    ):
        file_ir, file_results = analyse_single_file(
            """
            def sum_attrs(xs):
                return sum(
                    x.attr
                    for x in xs
                    if x.if_attr > 1
                    if x.another_attr == value
                )
            """
        )

        _, stderr = capfd.readouterr()
        assert helpers.stderr_matches(
            stderr,
            [helpers.as_warning("'value' potentially undefined")],
        )

        sum_attrs_symbol = Func("sum_attrs", interface=CallInterface(args=("xs",)))
        expected_file_ir = FileIr(
            context=make_root_context([sum_attrs_symbol], include_root_symbols=True),
            file_ir={
                sum_attrs_symbol: {
                    "gets": {
                        Name("xs"),
                        Name("x.attr", "x"),
                        Name("x.if_attr", "x"),
                        Name("x.another_attr", "x"),
                        Name("value"),
                    },
                    "sets": {
                        Name("x"),
                    },
                    "dels": set(),
                    "calls": {
                        Call(
                            name="sum",
                            args=CallArguments(args=("@GeneratorExp",)),
                            target=Builtin(name="sum"),
                        )
                    },
                }
            },
        )
        expected_file_results = FileResults(
            {
                sum_attrs_symbol.id: {
                    "gets": {"xs", "x.attr", "x.if_attr", "x.another_attr", "value"},
                    "sets": {"x"},
                    "dels": set(),
                    "calls": {"sum()"},
                }
            }
        )

        assert file_ir == expected_file_ir
        assert file_results == expected_file_results

    def test_generator_with_unbinding(
        self,
        analyse_single_file: Callable[[str], tuple[FileIr, FileResults]],
        make_root_context: MakeRootContextFn,
        builtin: Callable[[str], Builtin],
        capfd: pytest.CaptureFixture[str],
    ):
        file_ir, file_results = analyse_single_file(
            """
            def sum_attrs(xs, target):
                return sum(
                    x.attr
                    for x in xs
                    if x.if_attr > target.if_attr
                )

            def caller(ys, my_target):
                return sum_attrs(ys, my_target)
            """
        )

        _, stderr = capfd.readouterr()
        assert helpers.stdout_matches(stderr, [])

        symbols = {
            "sum_attrs": Func(
                "sum_attrs",
                interface=CallInterface(args=("xs", "target")),
            ),
            "caller": Func("caller", interface=CallInterface(args=("ys", "my_target"))),
        }
        expected_file_ir = FileIr(
            context=make_root_context(symbols.values(), include_root_symbols=True),
            file_ir={
                symbols["sum_attrs"]: {
                    "gets": {
                        Name("xs"),
                        Name("x.attr", "x"),
                        Name("x.if_attr", "x"),
                        Name("target.if_attr", "target"),
                    },
                    "sets": {
                        Name("x"),
                    },
                    "dels": set(),
                    "calls": {
                        Call(
                            "sum()",
                            args=CallArguments(args=("@GeneratorExp",)),
                            target=builtin("sum"),
                        ),
                    },
                },
                symbols["caller"]: {
                    "gets": {
                        Name("ys"),
                        Name("my_target"),
                        Name("my_target.if_attr", "my_target"),
                        Name("x.attr", "x"),
                        Name("x.if_attr", "x"),
                    },
                    "sets": {
                        Name("x"),
                    },
                    "dels": set(),
                    "calls": {
                        Call(
                            "sum_attrs()",
                            args=CallArguments(args=("ys", "my_target")),
                            target=symbols["sum_attrs"],
                        ),
                    },
                },
            },
        )
        expected_file_results = FileResults(
            {
                "sum_attrs": {
                    "gets": {"xs", "x.attr", "target.if_attr", "x.if_attr"},
                    "sets": {"x"},
                    "dels": set(),
                    "calls": {"sum()"},
                },
                "caller": {
                    "gets": {
                        "my_target.if_attr",
                        "x.if_attr",
                        "my_target",
                        "ys",
                        "x.attr",
                    },
                    "sets": {"x"},
                    "dels": set(),
                    "calls": {"sum_attrs()"},
                },
            }
        )

        assert file_ir == expected_file_ir
        assert file_results == expected_file_results

    def test_with_multiple_generators(
        self,
        analyse_single_file: Callable[[str], tuple[FileIr, FileResults]],
        make_root_context: MakeRootContextFn,
        builtin: Callable[[str], Builtin],
        capfd: pytest.CaptureFixture[str],
    ):
        file_ir, file_results = analyse_single_file(
            """
            def sum_attrs(wrapper):
                return sum(
                    x.a
                    for inner in wrapper.xss
                    if wrapper.xs > wrapper.xs_threshold
                    for x in inner.xs
                    if inner.x_value > inner.x_threshold
                )
            """
        )

        _, stderr = capfd.readouterr()
        assert helpers.stdout_matches(stderr, [])

        symbols = {
            "sum_attrs": Func("sum_attrs", interface=CallInterface(args=("wrapper",))),
        }
        expected_file_ir = FileIr(
            context=make_root_context(symbols.values(), include_root_symbols=True),
            file_ir={
                symbols["sum_attrs"]: {
                    "gets": {
                        Name("x.a", "x"),
                        Name("wrapper.xss", "wrapper"),
                        Name("wrapper.xs", "wrapper"),
                        Name("wrapper.xs_threshold", "wrapper"),
                        Name("inner.xs", "inner"),
                        Name("inner.x_value", "inner"),
                        Name("inner.x_threshold", "inner"),
                    },
                    "sets": {
                        Name("inner"),
                        Name("x"),
                    },
                    "dels": set(),
                    "calls": {
                        Call(
                            "sum()",
                            args=CallArguments(args=("@GeneratorExp",)),
                            target=builtin("sum"),
                        ),
                    },
                }
            },
        )
        expected_file_results = FileResults(
            {
                "sum_attrs": {
                    "gets": {
                        "inner.x_value",
                        "inner.x_threshold",
                        "wrapper.xs_threshold",
                        "inner.xs",
                        "x.a",
                        "wrapper.xss",
                        "wrapper.xs",
                    },
                    "sets": {"x", "inner"},
                    "dels": set(),
                    "calls": {"sum()"},
                }
            }
        )

        assert file_ir == expected_file_ir
        assert file_results == expected_file_results



"""CustomFunctionAnalyser for Python's collections."""
from __future__ import annotations

import ast
from typing import TYPE_CHECKING

from rattr.analyser.base import CustomFunctionAnalyser
from rattr.analyser.types import FunctionIr
from rattr.ast.util import fullname_of
from rattr.models.symbol import Call, CallArguments

if TYPE_CHECKING:
    from rattr.models.context import Context


def as_ast_lambda(expr: ast.expr) -> ast.Lambda:
    return ast.Lambda(
        args=ast.arguments(
            posonlyargs=[], args=[], kwonlyargs=[], kw_defaults=[], defaults=[]
        ),
        body=expr,
    )


class DefaultDictAnalyser(CustomFunctionAnalyser):
    @property
    def name(self) -> str:
        return "defaultdict"

    @property
    def qualified_name(self) -> str:
        return "collections.defaultdict"

    def on_def(
        self,
        name: str,
        node: ast.FunctionDef | ast.AsyncFunctionDef,
        ctx: Context,
    ) -> FunctionIr:
        return super().on_def(name, node, ctx)

    def on_call(self, name: str, node: ast.Call, ctx: Context) -> FunctionIr:
        # HACK Avoid circular import
        from rattr.analyser.function import FunctionAnalyser

        if len(node.args) == 0:
            return super().on_call(name, node, ctx)

        default_factory = node.args[0]

        # Three cases:
        #   1. A named callable         ->  Add call to name
        #   2. A lambda                 ->  Visit lambda
        #   3. Non-lambda expression    ->  Visit expression as dummy lambda

        if isinstance(default_factory, (ast.Name, ast.Attribute)):
            name = fullname_of(default_factory)
            target = ctx.get_call_target(name, culprit=node)
            results = FunctionIr.new(
                calls={
                    Call(
                        name=name,
                        args=CallArguments(),
                        target=target,
                        token=default_factory,
                    ),
                }
            )
        elif isinstance(default_factory, ast.Lambda):
            results = FunctionAnalyser(default_factory, ctx).analyse()
        else:
            results = FunctionAnalyser(as_ast_lambda(default_factory), ctx).analyse()

        return results



from __future__ import annotations

import operator
import sys
from itertools import takewhile
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Callable

    from rattr.versioning.typing import TypeAlias

    _Operator: TypeAlias = Callable[[str, str], bool]


_DOT_LEXEMES = (".",)
_OPERATOR_LEXEMES = (">", "<", "=")
_DIGIT_LEXEMES = ("1", "2", "3", "4", "5", "6", "7", "8", "9", "0")

_OPERATORS: dict[str, _Operator] = {
    ">=": operator.ge,
    ">": operator.gt,
    "<=": operator.le,
    "<": operator.lt,
    "=": operator.eq,
    "==": operator.eq,
    "!=": operator.ne,
}


def _version_cmp(op: _Operator, current: str, target: str) -> bool:
    _, current_minor_and_micro = current.split(".", maxsplit=1)
    _, target_minor_and_micro = target.split(".", maxsplit=1)
    return op(float(current_minor_and_micro), float(target_minor_and_micro))


def _parse_opcode(version: str) -> tuple[str, str]:
    opcode = "".join(takewhile(lambda c: c in _OPERATOR_LEXEMES, version))
    remainder = version[len(opcode) :]

    return opcode, remainder


def _parse_numeric(version: str) -> tuple[str, str]:
    numeric = "".join(takewhile(lambda c: c in _DOT_LEXEMES + _DIGIT_LEXEMES, version))
    remainder = version[len(numeric) :]

    return numeric, remainder


def _parse_version_string(raw: str) -> tuple[str, str]:
    if len(raw) < 1:
        raise ValueError("version must not be empty")

    _remainder = "".join(c for c in raw if not c.isspace())

    opcode, _remainder = _parse_opcode(_remainder)
    version, _remainder = _parse_numeric(_remainder)

    if not opcode:
        opcode = "=="

    if opcode not in _OPERATORS:
        raise ValueError(f"invalid operator: {opcode}")

    _parts = version.split(".")
    _parts_are_well_formed = all(p.isnumeric() for p in _parts)

    if (
        _remainder
        or not version
        or len(_parts) not in (2, 3)
        or not _parts_are_well_formed
        or not _parts[0] == "3"
    ):
        raise ValueError(f"malformed version: {raw}, expects '(<op>)?3.x(.y)?'")

    return opcode, version


def _current_version() -> tuple[int, int, int]:
    return sys.version_info.major, sys.version_info.minor, sys.version_info.micro


def _current_version_string(*, include_micro: bool = False) -> bool:
    major, minor, micro = _current_version()

    if include_micro:
        return f"{major}.{minor}.{micro}"

    return f"{major}.{minor}"


def is_python_version(version: str) -> bool:
    opcode, target_version = _parse_version_string(version)

    _include_micro = len(target_version.split(".")) == 3

    _operator = _OPERATORS[opcode]
    interpreter_version = _current_version_string(include_micro=_include_micro)

    return _version_cmp(_operator, interpreter_version, target_version)



from __future__ import annotations

from pathlib import Path

import pytest

from rattr.models.symbol import (
    AnyCallInterface,
    Builtin,
    Call,
    CallArguments,
    Class,
    Func,
    Import,
    Name,
)


@pytest.fixture()
def test_file() -> Path:
    return Path("test.py")


@pytest.fixture(autouse=True)
def __set_current_file(state, test_file):
    # Many symbols automatically derive a location
    # The derivation will error if the state's file is not set
    # Se just set it universally for all symbol tests.
    with state(current_file=test_file):
        yield


@pytest.fixture()
def simple_name() -> Name:
    return Name("jeffery")


@pytest.fixture()
def simple_builtin() -> Builtin:
    return Builtin("some_builtin_func")


@pytest.fixture()
def simple_import() -> Import:
    return Import("some_lib_func")


@pytest.fixture()
def simple_func() -> Func:
    return Func("some_user_func", interface=AnyCallInterface())


@pytest.fixture()
def simple_class() -> Class:
    return Class("SomeClass", interface=AnyCallInterface())


@pytest.fixture()
def simple_call() -> Call:
    return Call("my_func", args=CallArguments())



"""Rattr analyser util functions."""
from __future__ import annotations

import ast
import builtins
import hashlib
import io
import json
import re
import sys
from contextlib import redirect_stderr
from itertools import accumulate, chain, filterfalse
from os.path import isfile
from pathlib import Path
from string import ascii_lowercase
from time import perf_counter
from typing import TYPE_CHECKING

from isort.api import place_module

from rattr import error
from rattr.analyser.exc import RattrResultsError
from rattr.ast.types import AstComprehensions, AstLiterals, AstNodeWithName
from rattr.config import Config
from rattr.extra import DictChanges  # noqa: F401
from rattr.models.ir import FunctionIr
from rattr.models.symbol import (
    PYTHON_ATTR_ACCESS_BUILTINS,
    PYTHON_BUILTINS,
    Call,
    CallArguments,
    Class,
    Import,
    Name,
)
from rattr.module_locator.util import find_module_spec_fast

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable
    from typing import Any, Final, Type

    from rattr.analyser.types import RattrResults
    from rattr.ast.types import Identifier
    from rattr.models.context import Context
    from rattr.models.symbol import Symbol


re_rattr_name: Final = re.compile(r"^[A-Za-z_][\w\(\)\[\]\.]*$")


def get_basename_fullname_pair(
    node: ast.expr,
    safe: bool = False,
) -> tuple[str, str]:
    """Return the node's tuple of the basename and the fullname.

    If node (and it's children, recursively) are not StrictlyNameable, then a
    true name cannot be produced. If safe=True, then an attempt is made:

    #### All StrictlyNameable

        >>> get_basename_fullname_pair(
        ...     ast.parse("a.b[0].attr").body[0].value,
        ...     safe=True # or False
        ... )
        "a.b[].attr"

    #### Not all-StrictlyNameable, unsafe

        >>> get_basename_fullname_pair(
        ...     ast.parse("(a, b)[0].attr").body[0].value,
        ...     safe=False
        ... )
        error.RattrLiteralInNameable: ...

    #### Not all-StrictlyNameable, safe

        >>> get_basename_fullname_pair(
        ...     ast.parse("(a, b)[0].attr").body[0].value,
        ...     safe=True
        ... )
        "@Tuple[].attr"

    For the third example, the fact that `a.attr` is accessed is lost!

    NOTE Deprecated, see rattr.ast.util.names_of

    """
    config = Config()

    # Base case
    # ast.Name ⊂ StrictlyNameable
    if isinstance(node, ast.Name):
        return node.id, node.id

    # Recursive case
    # node ⊂ ( StrictlyNameable \ { ast.Name } )
    if isinstance(node, ast.Call):
        basename, sub_name = get_basename_fullname_pair(node.func, safe)
    elif isinstance(node, AstNodeWithName):
        basename, sub_name = get_basename_fullname_pair(node.value, safe)

    if isinstance(node, ast.Attribute):
        return basename, f"{sub_name}.{node.attr}"

    if isinstance(node, ast.Subscript):
        return basename, f"{sub_name}[]"

    if isinstance(node, ast.Call):
        if any(is_call_to(attr, node) for attr in PYTHON_ATTR_ACCESS_BUILTINS):
            return basename, ".".join(get_xattr_obj_name_pair(basename, node))

        return basename, f"{sub_name}()"

    if isinstance(node, ast.Starred):
        return basename, f"*{sub_name}"

    # Error case
    # node ⊂ Nameable ^ node ⊄ StrictlyNameable
    if safe:
        return (
            f"{config.LITERAL_VALUE_PREFIX}{node.__class__.__name__}",
            f"{config.LITERAL_VALUE_PREFIX}{node.__class__.__name__}",
        )

    _error_class: Type[TypeError] = TypeError
    if isinstance(node, ast.UnaryOp):
        _error_class = error.RattrUnaryOpInNameable
    elif isinstance(node, ast.BinOp):
        _error_class = error.RattrBinOpInNameable
    elif isinstance(node, ast.Constant):
        _error_class = error.RattrConstantInNameable
    elif isinstance(node, AstLiterals):
        _error_class = error.RattrLiteralInNameable
    elif isinstance(node, AstComprehensions):
        _error_class = error.RattrComprehensionInNameable
    elif isinstance(node, ast.GeneratorExp):
        _error_class = error.RattrComprehensionInNameable
    else:
        _error_class = TypeError

    raise _error_class(f"line {node.lineno}: {ast.dump(node)}")


def get_basename(node: ast.expr, safe: bool = False) -> str:
    """Return the `_identifier` of the innermost ast.Name node.

    NOTE Deprecated, see rattr.ast.utils
    """
    return get_basename_fullname_pair(node, safe)[0]


def get_fullname(node: ast.expr, safe: bool = False) -> str:
    """Return the fullname of the given node.

    NOTE Deprecated, see rattr.ast.utils
    """
    return get_basename_fullname_pair(node, safe)[1]


def get_attrname(node: ast.expr) -> str:
    """Return the `_identifier` of the outermost attribute in a nested name.

    E.g. `a.b.c.d` -> `d`
    >>> attr = ast.parse("object.attribute").body[0].value
    """
    if isinstance(node, ast.Name):
        return node.id

    if isinstance(node, ast.Attribute):
        return node.attr

    if isinstance(node, ast.Call):
        return get_attrname(node.func)

    raise TypeError(f"line {node.lineno}: {ast.dump(node)}")


def unravel_names(
    node: ast.expr,
    *,
    get_name: Callable[[ast.expr], str] = get_basename,
) -> Iterable[str]:
    """Return the basename of each nameable in the given node.

    >>> ravelled_names = ast.parse("a, b = 1, 2").body[0].targets[0]
    >>> list(unravel_names(ravelled_names))
    ["a", "b"]

    >>> ravelled_names = ast.parse("(a, b), c, d.e = ...").body[0].targets[0]
    >>> list(unravel_names(ravelled_names))
    ["a", "b", "c", "d"]

    The name getter can be overridden for example as `get_fullname`:
    >>> ravelled_names = ast.parse("a.attr = 1").body[0].targets[0]
    >>> list(unravel_names(ravelled_names))
    ["a"]
    >>> list(unravel_names(ravelled_names, get_name=get_fullname))
    ["a.attr"]

    NOTE Deprecated, see rattr.ast.util.unravel_names

    """
    if isinstance(node, AstNodeWithName):
        return [get_name(node)]

    if isinstance(node, (ast.Tuple, ast.List)):
        ravelled = [unravel_names(i, get_name=get_name) for i in node.elts]
        return chain.from_iterable(ravelled)

    raise TypeError(f"line {node.lineno}: {ast.dump(node)}")


def is_call_to(target: str, node: ast.Call) -> bool:
    """Return `True` if the given node is a __direct__ call to `target`.

    In the case `getattr(o, a).some_call()`, the basename of `.some_call()` is
    `getattr` -- this is the correct basename but gives incorrect behaviour;
    the args of `getattr` are not properly resolved and `.some_call()` is not
    fully visited.

    Thus rather than determining if a call is to `getattr` by checking the
    basename against "getattr", use `is_call_to("getattr", node)`.

    NOTE Deprecated, see rattr.ast.utils

    """
    if not isinstance(node, ast.Call):
        raise TypeError(f"line {node.lineno}: {ast.dump(node)}")

    if not isinstance(node.func, ast.Name):
        return False

    return node.func.id == target


def has_affect(builtin: str) -> bool:
    """Return `True` if the given builtin can affect accessed attributes."""
    if builtin not in PYTHON_BUILTINS:
        raise ValueError(f"'{builtin}' is not a Python builtin")

    return builtin in PYTHON_ATTR_ACCESS_BUILTINS


def get_xattr_obj_name_pair(
    xattr: str,
    node: ast.Call,
    warn: bool = False,
) -> tuple[str, str]:
    """Return the object-name pair for a call to getattr, setattr, etc.

    NOTE Deprecated, see rattr.ast.utils
    """
    if len(node.args) < 2:
        error.fatal(f"invalid call to '{xattr}', not enough args", node)

    this = f"'{xattr}'"
    obj, attr = node.args[0], node.args[1]

    # NOTE Only warn if told to, avoids repeated warnings
    if warn and not isinstance(attr, ast.Str):
        error.error(f"{this} expects name to be a string literal", node)

    if isinstance(attr, ast.Str):
        attr_name = attr.s
    else:
        attr_name = f"<{get_fullname(attr, safe=True)}>"

    # Nested calls must be to the same function
    if isinstance(obj, ast.Call) and not is_call_to(xattr, obj):
        error.fatal(f"{this} object must be a name or a call to {this}", node)

    # Recurive Case
    # E.g. `getattr(getattr(a, 'b'), 'c')`
    if isinstance(obj, ast.Call):
        return ".".join(get_xattr_obj_name_pair(xattr, obj)), attr_name

    # Base Case
    # NOTE Call ∈ AstStrictlyNameable, thus base case comes second
    if isinstance(obj, AstNodeWithName):
        return get_fullname(node.args[0]), attr_name

    raise TypeError(f"line {node.lineno}: {ast.dump(node)}")


def get_decorator_name(decorator: ast.expr) -> str:
    """Return name of the given decorator.."""
    if not isinstance(decorator, (ast.Name, ast.Attribute, ast.Call)):
        raise TypeError("Decorator must be an ast.Name, ast.Attribute, or ast.Call")

    return get_basename(decorator)


def get_nth_argument_name(arguments: ast.arguments, n: int) -> str:
    """Return the name of the nth positional argument, or Ɛ if out-of-range."""
    _arguments = arguments.args

    if len(_arguments) == 0:
        return ""

    if not (0 <= n < len(_arguments)):
        return ""

    return _arguments[n].arg


def get_first_argument_name(arguments: ast.arguments) -> str:
    return get_nth_argument_name(arguments, 0)


def get_second_argument_name(arguments: ast.arguments) -> str:
    return get_nth_argument_name(arguments, 1)


def has_annotation(
    name: str,
    target: ast.FunctionDef | ast.AsyncFunctionDef | ast.ClassDef,
) -> bool:
    """Return `True` if the function is decorated with the given annotation."""
    return name in map(get_attrname, target.decorator_list)


def get_annotation(
    name: str,
    target: ast.FunctionDef | ast.AsyncFunctionDef | ast.ClassDef,
) -> ast.expr | None:
    """Return the decorator node for the given annotation on the function."""
    matching: list[ast.expr] = []

    for decorator in target.decorator_list:
        suffix = get_attrname(decorator)

        if suffix != name:
            continue

        matching.append(decorator)

    if len(matching) < 1:
        return None
    if len(matching) > 1:
        error.fatal(
            f"duplicated annotation {name!r} on {target.name!r}",
            culprit=target,
        )

    return matching[0]


def safe_eval(expr: ast.expr, culprit: ast.AST) -> Any | None:
    """Return the given expression as evaluated at compile-time.

    NOTE
        Use this instead of Python's `eval` to avoid potential arbitrary code
        execution.

    """
    _error = "unable to evaluate {} at compile-time"

    if isinstance(expr, ast.Num):
        return expr.n

    if isinstance(expr, (ast.Str, ast.Bytes)):
        return expr.s

    if isinstance(expr, ast.NameConstant):
        return expr.value

    if isinstance(expr, ast.List):
        return [safe_eval(e, culprit) for e in expr.elts]

    if isinstance(expr, ast.Tuple):
        return tuple([safe_eval(e, culprit) for e in expr.elts])

    if isinstance(expr, ast.Set):
        return set([safe_eval(e, culprit) for e in expr.elts])

    if isinstance(expr, ast.Dict):
        if not all(k is not None for k in expr.keys):
            return error.fatal(_error.format("dictionary unpacking"), culprit)

        keys = map(lambda k: safe_eval(k, culprit), expr.keys)
        values = map(lambda v: safe_eval(v, culprit), expr.values)

        return {k: v for k, v in zip(keys, values)}

    return error.fatal(_error.format(f"'{ast.dump(expr)}'"), culprit)


def parse_annotation(
    name: str,
    fn_def: ast.FunctionDef | ast.AsyncFunctionDef,
) -> tuple[list[Any], dict[str, Any]]:
    """Return the positional and keyword arguments of the annotation."""
    annotation = get_annotation(name, fn_def)

    pos_args: list[Any] = []
    named_args: dict[str, Any] = {}

    if not annotation:
        return pos_args, named_args

    if not isinstance(annotation, ast.Call):
        return pos_args, named_args

    for arg in annotation.args:
        pos_args.append(safe_eval(arg, fn_def))

    for kwarg in annotation.keywords:
        named_args[kwarg.arg] = safe_eval(kwarg.value, fn_def)

    return pos_args, named_args


def is_name(target: str | object) -> bool:
    """Return `True` if the given name is a valid Python identifier."""
    if not isinstance(target, str):
        return False

    name = target.removeprefix("*").removeprefix("@")
    return re_rattr_name.fullmatch(name) is not None


def is_set_of_names(target: set[str] | object) -> bool:
    return isinstance(target, set) and all(is_name(name) for name in target)


def is_list_of_names(target: list[str] | object) -> bool:
    return isinstance(target, list) and all(is_name(name) for name in target)


def is_list_of_call_specs(
    call_specs: list[tuple[Identifier, tuple[list[Identifier], dict[Identifier, str]]]]
    | object,
) -> bool:
    if not isinstance(call_specs, list):
        return False

    for spec in call_specs:
        if not isinstance(spec, tuple):
            return False

        if len(spec) != 2:
            return False

        target_name, target_args = spec

        if not is_name(target_name):
            return False

        if not isinstance(target_args, tuple):
            return False

        if len(target_args) != 2:
            return False

        target_pos_args, target_keyword_args = target_args

        if not is_list_of_names(target_pos_args):
            return False

        for arg_name, local_identifier in target_keyword_args.items():
            if not is_name(arg_name):
                return False
            if not is_name(local_identifier):
                return False

    return True


def validate_rattr_results(rattr_results: RattrResults) -> None:
    """Raise a RattrResultsError if the given results are invalid.

    Raises:
        RattrResultsError: The given rattr results are invalid.
    """
    for key in ("gets", "sets", "dels"):
        if not is_set_of_names(rattr_results[key]):
            raise RattrResultsError(
                f"'rattr_results' expects a set[Identifier] for {key!r}, where "
                f"Identifier is a type alias for str"
            )

    if not is_list_of_call_specs(rattr_results["calls"]):
        raise RattrResultsError(
            "'rattr_results' expects 'calls' to be a "
            "list[tuple[TargetName, tuple[list[PositionalArgumentName], "
            "dict[KeywordArgumentName, LocalIdentifier]]]]; "
            "where TargetName, PositionalArgumentName, KeywordArgumentName, and "
            "LocalIdentifier are type aliases for str"
        )


def parse_rattr_results_from_annotation_args_impl(
    fn_def: ast.FunctionDef | ast.AsyncFunctionDef,
) -> RattrResults:
    stderr_stream = io.StringIO()
    has_likely_missing_comma = False

    try:
        with redirect_stderr(stderr_stream):
            decorator_args, decorator_kwargs = parse_annotation("rattr_results", fn_def)
    except SystemExit as exc:
        stderr = stderr_stream.getvalue()

        # TODO On switching to proper logging use logger here
        for line in stderr.splitlines():
            if "unable to evaluate" in line:
                has_likely_missing_comma = True
            else:
                print(line)

        if has_likely_missing_comma:
            error.fatal(
                "unable to parse 'rattr_results', you are likely missing a "
                "comma in 'calls'",
                culprit=fn_def,
            )

        raise exc

    if decorator_args != []:
        error.fatal(
            f"unexpected positional arguments to 'rattr_results'; expected "
            f"none, got {decorator_args}",
            culprit=fn_def,
        )

    rattr_results_from_annotation: RattrResults = {
        "gets": set(),
        "sets": set(),
        "dels": set(),
        "calls": list(),
    }
    for key, results in decorator_kwargs.items():
        if key not in rattr_results_from_annotation:
            continue
        rattr_results_from_annotation[key] = results

    if decorator_kwargs.keys() - rattr_results_from_annotation.keys():
        error.fatal(
            f"unexpected keyword arguments to 'rattr_results'; expected any of "
            f"{list(rattr_results_from_annotation.keys())}, got "
            f"{list(decorator_kwargs.keys())}",
            culprit=fn_def,
        )

    return rattr_results_from_annotation


def parse_rattr_results_from_annotation(
    fn_def: ast.FunctionDef | ast.AsyncFunctionDef,
    *,
    context: Context,
) -> FunctionIr:
    """Return the IR for the given function, assuming it is annotated."""
    rattr_results = parse_rattr_results_from_annotation_args_impl(fn_def)

    try:
        validate_rattr_results(rattr_results)
    except RattrResultsError as exc:
        error.fatal(exc.message, culprit=fn_def)

    def as_name(name: Identifier) -> Name:
        return Name(
            name=name,
            basename=name.replace("*", "").split(".")[0],
            token=fn_def,
        )

    def as_call(
        call: tuple[Identifier, tuple[list[Identifier], dict[Identifier, Identifier]]],
    ) -> Call:
        (target_name, (args, kwargs)) = call
        return Call(
            name=target_name,
            args=CallArguments(args=args, kwargs=kwargs),
            target=context.get_call_target(target_name, culprit=fn_def),
            token=fn_def,
        )

    return {
        "gets": {as_name(name) for name in rattr_results["gets"]},
        "sets": {as_name(name) for name in rattr_results["sets"]},
        "dels": {as_name(name) for name in rattr_results["dels"]},
        "calls": {as_call(call) for call in rattr_results["calls"]},
    }


def is_blacklisted_module(module: str) -> bool:
    """Return `True` if the given module matches a blacklisted pattern.

    NOTE Deprecated, see rattr.module_locator.util.is_in_import_blacklist
    """
    config = Config()

    # Exclude stdlib modules such as the built-in "_thread"
    if is_stdlib_module(module):
        return False

    return any(re.fullmatch(p, module) for p in config.blacklist_patterns)


def is_pip_module(module: str) -> bool:
    """Return `True` if the given module is pip installed.

    NOTE Deprecated, see rattr.ast.place
    """
    pip_install_locations = (".+/site-packages.*",)

    spec = find_module_spec_fast(module)

    if spec is None or spec.origin is None:
        return False

    return any(
        re.fullmatch(
            pip_install_location,
            # No backslashes, bad windows!
            spec.origin.replace("\\", "/"),
        )
        for pip_install_location in pip_install_locations
    )


def is_stdlib_module(module: str) -> bool:
    """Return `True` if the given module is in the Python standard library.

    >>> is_stdlib_module("math")
    True

    >>> is_stdlib_module("math.pi")
    True

    NOTE Deprecated, see rattr.ast.place
    """
    return place_module(module) == "STDLIB"


def is_in_builtins(name_or_qualified_name: str) -> bool:
    return name_or_qualified_name in dir(builtins)


class timer:
    """Context manager to return the context's elapsed time in seconds."""

    def __enter__(self):
        self.start = perf_counter()
        self.end = 0.0
        return self

    def __exit__(self, *_):
        self.end = perf_counter()

    @property
    def time(self):
        return self.end - self.start


class read:
    """Context manager to return file contents and the number of lines."""

    def __init__(self, file: str) -> None:
        self.file = file

    def __enter__(self) -> tuple[int, str]:
        with open(self.file, "r") as f:
            lines = f.readlines()
        return len(lines) + 1, "".join(lines)

    def __exit__(self, *_) -> None:
        pass


def get_starred_imports(
    symbols: list[Symbol],
    seen: Iterable[Import],
) -> list[Import]:
    """Return the starred imports in the given symbols."""
    starred: list[Import] = []

    for symbol in symbols:
        if not isinstance(symbol, Import):
            continue

        if symbol.name != "*":
            continue

        if symbol in seen:
            continue

        starred.append(symbol)

    return starred


def get_function_body(
    node: ast.Lambda | ast.FunctionDef | ast.AsyncFunctionDef,
) -> list[ast.stmt]:
    """Return the body of the given function as a list of statements."""
    if isinstance(node, ast.Lambda):
        return [node.body]

    if isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef)):
        return node.body

    raise TypeError(f"line {node.lineno}: {ast.dump(node)}")


def get_assignment_targets(
    node: ast.Assign | ast.AnnAssign | ast.AugAssign | ast.NamedExpr,
) -> list[ast.expr]:
    """Return the given assignment's targets as a list."""
    if isinstance(node, ast.Assign):
        return node.targets

    if isinstance(node, (ast.AnnAssign, ast.AugAssign, ast.NamedExpr)):
        return [node.target]

    raise TypeError(f"line {node.lineno}: {ast.dump(node)}")


def get_contained_walruses(
    node: ast.Assign | ast.AnnAssign | ast.AugAssign | ast.NamedExpr,
) -> list[ast.NamedExpr]:
    """Return the walruses in the RHS of the given assignment.

    >>> get_nested_walruses(ast.parse("a = (b := c)"))
    [NamedExpr(target=Name(id='b', ctx=Store()), value=Name(id='c', ctx=Load())), ]

    >>> get_nested_walruses(ast.parse("a = (b, c := d)"))
    [NamedExpr(target=Name(id='c', ctx=Store()), value=Name(id='d', ctx=Load())), ]
    """
    if not walrus_in_rhs(node):
        return list()

    if isinstance(node.value, (ast.Tuple, ast.List)):
        rhs_values = node.value.elts
    else:
        rhs_values = [node.value]

    return list(filter(lambda v: isinstance(v, ast.NamedExpr), rhs_values))


def assignment_is_one_to_one(
    node: ast.Assign | ast.AnnAssign | ast.AugAssign | ast.NamedExpr,
) -> bool:
    """Return `True` if the given assignment is one-to-one.

    NOTE Deprecated, see rattr.ast.util.assignment_is_one_to_one
    """
    _iterable = (ast.Tuple, ast.List)

    targets = get_assignment_targets(node)

    # Ensure LHS is singular
    if len(targets) > 1 or any(isinstance(t, _iterable) for t in targets):
        return False

    return not isinstance(node.value, _iterable)


def lambda_in_rhs(
    node: ast.Assign | ast.AnnAssign | ast.AugAssign | ast.NamedExpr,
) -> bool:
    """Return `True` if the RHS of the given assignment contains a lambda.

    NOTE Deprecated, see rattr.ast.util.has_lambda_in_rhs
    """
    _iterable = (ast.Tuple, ast.List)

    if isinstance(node.value, ast.Lambda):
        return True
    elif isinstance(node.value, _iterable):
        return any(isinstance(v, ast.Lambda) for v in node.value.elts)
    else:
        return False


def walrus_in_rhs(
    node: ast.Assign | ast.AnnAssign | ast.AugAssign | ast.NamedExpr,
) -> bool:
    """Return `True` if the RHS contains a walrus operator.

    NOTE Deprecated, see rattr.ast.util.has_walrus_in_rhs
    """
    _iterable = (ast.Tuple, ast.List)

    if isinstance(node.value, ast.NamedExpr):
        return True
    elif isinstance(node.value, _iterable):
        return any(isinstance(v, ast.NamedExpr) for v in node.value.elts)
    else:
        return False


def namedtuple_in_rhs(
    node: ast.Assign | ast.AnnAssign | ast.AugAssign | ast.NamedExpr,
) -> bool:
    """Return `True` if the RHS contains a namedtuple creation.

    NOTE Deprecated, see rattr.ast.util.has_namedtuple_declaration_in_rhs
    """
    _iterable = (ast.Tuple, ast.List)

    def _target_is_namedtuple(call: ast.Call) -> bool:
        # HACK
        #   This is a naive approach, when enum / class-namedtuple / etc detection is
        #   refactored we can do this a bit better.
        #   See: cls.py::is_enum, cls.py::is_namedtuple
        name = get_fullname(call.func, safe=True)
        return name == "namedtuple" or name.endswith(".namedtuple")

    if isinstance(node.value, ast.Call):
        return _target_is_namedtuple(node.value)
    elif isinstance(node.value, _iterable):
        return any(
            _target_is_namedtuple(value)
            for value in node.value.elts
            if isinstance(value, ast.Call)
        )
    else:
        return False


def _attrs_from_list_of_strings(attrs_argument: ast.List) -> list[str]:
    """NOTE Deprecated, see rattr.ast.util.unpack_ast_list_of_strings."""
    attrs: list[str] = [
        arg.value
        for arg in attrs_argument.elts
        if isinstance(arg, ast.Constant)
        if isinstance(arg.value, str)
    ]

    if len(attrs) != len(attrs_argument.elts):
        raise SyntaxError

    return attrs


def _attrs_from_space_delimited_string(attrs_argument: ast.Constant) -> list[str]:
    """NOTE Deprecated, see rattr.ast.util.parse_space_delimited_ast_string."""
    if not isinstance(attrs_argument.value, str):
        raise SyntaxError

    attrs = attrs_argument.value.split(" ")

    if not all(attr.isidentifier() for attr in attrs):
        raise SyntaxError

    return attrs


def _namedtuple_attrs_from_second_argument(attrs_argument: ast.AST) -> list[str]:
    """NOTE Deprecated."""
    if isinstance(attrs_argument, ast.List):
        return _attrs_from_list_of_strings(attrs_argument)

    if isinstance(attrs_argument, ast.Constant):
        return _attrs_from_space_delimited_string(attrs_argument)

    raise SyntaxError


def get_namedtuple_attrs_from_call(
    node: ast.Assign | ast.AnnAssign | ast.AugAssign | ast.NamedExpr,
) -> tuple[list[str], dict[str, str]]:
    """Return the args/attrs of the namedtuple constructed by this assignment by call.

    #### Note
    Assumes one-to-one assignment.

    #### Examples
    >>> node = ast.parse("p = namedtuple('p', ['x', 'y'])").body[0].value
    >>> get_namedtuple_init_signature_from_assignment(node)
    ["self", "x", "y]

    Raises:
        TypeError: The node is not an assignment of a call.

    Returns:
        tuple[list[str], dict[str, str]]: The positional and keyword args of the init.

    NOTE Deprecated, see rattr.ast.util.namedtuple_init_signature_from_declaration
    """
    _invalid_signature_error = (
        "namedtuple expects exactly two positional arguments (i.e. name, attrs)"
    )
    _invalid_second_parameter_value_error = (
        "namedtuple expects the second positional argument to be a list of valid "
        "identifiers as either a list of string-literals or a space-delimited "
        "string-literal"
    )

    # Parse call arguments
    if not isinstance(node.value, ast.Call):
        raise TypeError

    namedtuple_call_arguments = node.value.args

    if len(namedtuple_call_arguments) != 2:
        error.fatal(_invalid_signature_error, culprit=node.value)

    _, namedtuple_attrs_argument = namedtuple_call_arguments

    try:
        attrs = _namedtuple_attrs_from_second_argument(namedtuple_attrs_argument)
    except SyntaxError:
        error.fatal(_invalid_second_parameter_value_error, culprit=node.value)

    return ["self", *attrs]


def class_in_rhs(
    node: ast.Assign | ast.AnnAssign | ast.AugAssign | ast.NamedExpr,
    context: Context,
) -> bool:
    """Return `True` if the RHS of the given assignment is a class init."""
    _iterable = (ast.Tuple, ast.List)

    def expr_is_class(expr):
        full = get_fullname(expr, safe=True)

        try:
            target = context.get_call_target(full, expr, warn=False)
        except ValueError:
            return False

        return isinstance(target, Class)

    if isinstance(node.value, ast.Call):
        return expr_is_class(node.value)
    elif isinstance(node.value, _iterable):
        return any(expr_is_class(e) for e in node.value.elts)
    else:
        return False


def is_starred_import(node: ast.Import | ast.ImportFrom) -> bool:
    """Return `True` if the given import is a `from module import *`.

    NOTE Deprecated, see rattr.ast.util.is_starred_import
    """
    if not isinstance(node, ast.ImportFrom):
        return False

    return any(target.name == "*" for target in node.names)


def is_relative_import(node: ast.Import | ast.ImportFrom) -> bool:
    """Return `True` if the given import is a relative import.

    NOTE Deprecated, see rattr.ast.util.is_relative_import
    """
    if not isinstance(node, ast.ImportFrom):
        return False

    return node.level != 0


def module_name_from_file_path(file: Path | str | None) -> str | None:
    """Return the recognised name of the given module.

    NOTE Deprecated, see rattr.module_locator.util.derive_module_name_from_path
    """
    if file is None:
        raise ValueError

    if isinstance(file, Path):
        file = str(file)

    module = file.replace("/", ".").replace("\\", ".")

    if module.endswith(".__init__.py"):
        module = module[:-12]

    if module.endswith(".py"):
        module = module[:-3]

    module = module.strip(".")

    # Get possible module names in relevancy order
    parts = module.split(".")
    dotted_parts = list(chain.from_iterable(zip(parts, ["."] * len(parts))))
    possible = list(accumulate(reversed(dotted_parts), lambda p0, p1: f"{p1}{p0}"))
    well_formed = list(filterfalse(lambda p: p.startswith("."), possible))
    well_formed = list(wf[:-1] for wf in well_formed if len(wf) >= 1)
    ordered = list(reversed(well_formed))

    for p in [*ordered, None]:
        if find_module_spec_fast(p) is None:
            break

    return p


def get_absolute_module_name(base: str, level: int, target: str) -> str:
    """Return the absolute import for the given relative import."""
    config = Config()

    if config.state.current_file.name == "__init__.py":
        level -= 1

    if level > 0:
        new_base = ".".join(base.split(".")[:-level])
    else:
        new_base = base

    return f"{new_base}.{target}"


def get_function_form(node: ast.FunctionDef | ast.AsyncFunctionDef) -> str:
    """Return the string representing the function form.

    Given the function, fn:
        ```
        def fn(a, b, c=None, *args, **kwargs):
            # ...
        ```

    The form is:
        `fn(a, b, c, *args, **kwargs)`

    """
    arguments = [a.arg for a in node.args.args]

    if node.args.vararg is not None:
        arguments.append(f"*{node.args.vararg.arg}")

    if node.args.kwarg is not None:
        arguments.append(f"**{node.args.kwarg.arg}")

    return f"'{node.name}({', '.join(arguments)})'"


def is_excluded_name(name: str) -> bool:
    """Return `True` if the given name is in the exclude list."""
    config = Config()
    return any(
        pattern.fullmatch(name) is not None
        for pattern in config.arguments.re_excluded_names
    )


def is_method_on_constant(name: str) -> bool:
    """Return `True` if the name is a call to method on a constant."""
    if sys.version_info.major == 3 and sys.version_info.minor <= 7:
        return name.startswith(
            (
                "@Num.",
                "@Str.",
                "@Bytes.",
                "@NameConstant.",
            )
        )

    if sys.version_info.major == 3 and sys.version_info.minor >= 8:
        return name.startswith("@Constant.")

    raise NotImplementedError


def is_method_on_cast(name: str) -> bool:
    """Return `True` if the name is a call to a method on a cast."""
    # NOTE All builtin functions return primitives (often `None`), ...
    builtins = {b for b in PYTHON_BUILTINS if b.startswith((*ascii_lowercase,))}

    # NOTE ... with the following exceptions
    builtins = builtins.difference(
        {
            "eval",
            "exec",
            "getattr",
            "iter",
            "next",
            "slice",
            "super",
            "type",
        }
    )

    # NOTE
    #   Direct call to `<primitive_returner>.method_name()`, do not allow call
    #   to call, etc because one may return non-primitive
    patterns = tuple((f"{b}\\(\\)\\.[^\\(\\)\\[\\]\\.]+" for b in builtins))

    return any(re.fullmatch(p, name) for p in patterns)


def is_method_on_primitive(name: str) -> bool:
    """Return `True` if the given name is a method on a primitive type."""
    if is_method_on_constant(name):
        return True

    if is_method_on_cast(name):
        return True

    return False


def get_dynamic_name(fn_name: str, node: ast.Call, pattern: str) -> Name:
    """Return the `Name` of the given pattern constructed from the call.

    Where first is the resolved name of the first call argument, and second is
    the resolved name of the second call argument.

    Assumes that the first argument is a variable (or nested call) and the
    second argument is a string -- names are resolved as such.

    Allows nested calls to functions with the same name (i.e. `name`).

    >>> call = ast.parse("getattr(object, 'attribute')").body[0].value
    >>> get_dynamic_name("getattr", call, "{first}.{second}")
    Name("object.attribute", "object")

    >>> call = ast.parse("getattr(getattr(a, 'b'), 'c')").body[0].value
    >>> get_dynamic_name("getattr", call, "{first}.{second}")
    Name("a.b.c", "a")

    >>> call = ast.parse("getattr(another_fn(a, 'b'), 'c')").body[0].value
    >>> get_dynamic_name("getattr", call, "{first}.{second}")
    sys.exit(1)

    >>> call = ast.parse("get_sub_attr(obj, 'attr')").body[0].value
    >>> get_dynamic_name("get_sub_attr", call, "{first}.sub.{second}")
    Name("obj.sub.attr", "obj")

    >>> call = ast.parse("get_sub_attr(get_sub_attr(o, 'in'), 'out')").body[0].value
    >>> get_dynamic_name("get_sub_attr", call, "{first}.mid.{second}")
    Name("o.in.mid.out", "o")

    >>> call = ast.parse("get_sub(obj)").body[0].value
    >>> get_dynamic_name("get_sub", call, "{first}.sub")
    Name("obj.sub", "obj")

    """
    first, second = get_xattr_obj_name_pair(fn_name, node, warn=True)
    basename = first.split(".")[0].replace("*", "").replace("[]", "").replace("()", "")

    return Name(
        name=pattern.format(first=first, second=second),
        basename=basename,
        token=node,
    )


def get_file_hash(filepath: str, blocksize: int = 2**20) -> str:
    """Return the hash of the given file, with a default blocksize of 1MiB."""
    _hash = hashlib.md5()

    if not isfile(filepath):
        return _hash

    with open(filepath, "rb") as f:
        while True:
            buffer = f.read(blocksize)

            if not buffer:
                break

            _hash.update(buffer)

    return _hash.hexdigest()


def cache_is_valid(filepath: str, cache_filepath: str) -> bool:
    """Return `True` if the cache has the correct hash."""
    if not isfile(filepath):
        return False

    if isfile(cache_filepath):
        with open(cache_filepath, "r") as f:
            cache: dict[str, Any] = json.load(f)
    else:
        return False

    received_hash = cache.get("filehash", None)
    expected_hash = get_file_hash(filepath)

    if filepath != cache.get("filepath", None):
        return False

    if received_hash != expected_hash:
        return False

    # Check imports
    for _import in cache.get("imports", {}):
        file, hash = _import["filepath"], _import["filehash"]
        if hash != get_file_hash(file):
            return False

    return True
